// Copyright (c) 1999, A.H. van den Boogaard

#include "convert.h"
#include "myexcept.h"

using FeaTure::Vector;
using FeaTure::CollectArray;
using FeaTure::StressType;
using FeaTure::gen;
using FeaTure::psn;
using FeaTure::pss;
using FeaTure::axi;
using FeaTure::uni;

Vector Tensor2StrainVector( const Tensor2Sym &eps, bool expanded, StressType st )
{
   int size;
   size = expanded ? ExpandedStressSize(st) : StressSize(st);
   Vector strain;
   strain.resize( size );

   if ( expanded )
   {
      switch ( st )
      {
         case gen: strain(5) = 2*eps(2,3);
                   strain(6) = 2*eps(3,1);
         case psn:
         case pss:
         case axi: strain(4) = 2*eps(1,2);
         case uni: strain(1) = eps(1,1);
                   strain(2) = eps(2,2);
                   strain(3) = eps(3,3);
                   break;
         default:  throw featured_error( "undefined stress type in Tensor2StrainVector" );
      }
   }
   else
   {
      switch ( st )
      {
         case gen: strain(5) = 2*eps(2,3);
                   strain(6) = 2*eps(3,1);
         case axi: strain(4) = 2*eps(1,2);
                   strain(3) = eps(3,3);
                   strain(2) = eps(2,2);
                   strain(1) = eps(1,1);
                   break;
         case psn:
         case pss: strain(3) = 2*eps(1,2);
                   strain(2) = eps(2,2);
         case uni: strain(1) = eps(1,1);
                   break;
         default:  throw featured_error( "undefined stress type in Tensor2StrainVector" );
      }
   }
   return strain;
}

Vector Tensor2StressVector( const Tensor2Sym &sig, bool expanded, StressType st )
{
   int size;
   size = expanded ? ExpandedStressSize(st) : StressSize(st);
   Vector stress;
   stress.resize( size );

   if ( expanded )
   {
      switch ( st )
      {
         case gen: stress(5) = sig(2,3);
                   stress(6) = sig(3,1);
         case psn:
         case pss:
         case axi: stress(4) = sig(1,2);
         case uni: stress(1) = sig(1,1);
                   stress(2) = sig(2,2);
                   stress(3) = sig(3,3);
                   break;
         default:  throw featured_error( "undefined stress type in Tensor2StressVector" );
      }
   }
   else
   {
      switch ( st )
      {
         case gen: stress(5) = sig(2,3);
                   stress(6) = sig(3,1);
         case axi: stress(4) = sig(1,2);
                   stress(3) = sig(3,3);
                   stress(2) = sig(2,2);
                   stress(1) = sig(1,1);
                   break;
         case psn:
         case pss: stress(3) = sig(1,2);
                   stress(2) = sig(2,2);
         case uni: stress(1) = sig(1,1);
                   break;
         default:  throw featured_error( "undefined stress type in Tensor2StressVector" );
      }
   }
   return stress;
}

Tensor2Sym StressVector2Tensor( const Vector &stress, bool expanded, StressType st )
{
   Tensor2Sym sig;
   if ( st != gen ) sig = 0;

   if ( expanded )
   {
      switch ( st )
      {
         case gen: sig(3,1) = stress(6);
                   sig(3,2) = stress(5);
         case axi:
         case psn:
         case pss: sig(2,1) = stress(4);
         case uni: sig(1,1) = stress(1);
                   sig(2,2) = stress(2);
                   sig(3,3) = stress(3);
                   break;
         default:  throw featured_error( "undefined stress type in StressVector2Tensor" );
      }
   }
   else
   {
      switch( st )
      {
         case gen: sig(3,1) = stress(6);
                   sig(3,2) = stress(5);
         case axi: sig(2,1) = stress(4);
                   sig(1,1) = stress(1);
                   sig(2,2) = stress(2);
                   sig(3,3) = stress(3);
                   break;
         case psn:
         case pss: sig(2,1) = stress(3);
                   sig(2,2) = stress(2);
         case uni: sig(1,1) = stress(1);
                   break;
         default:  throw featured_error( "undefined stress type in StressVector2Tensor" );
      }
   }

   return sig;
}

Tensor2Sym StrainVector2Tensor( const Vector &strain, bool expanded, StressType st )
{
   Tensor2Sym eps;
   if ( st != gen ) eps = 0;

   if ( expanded )
   {
      switch ( st )
      {
         case gen: eps(3,1) = 0.5*strain(6);
                   eps(3,2) = 0.5*strain(5);
         case axi:
         case psn:
         case pss: eps(2,1) = 0.5*strain(4);
         case uni: eps(1,1) = strain(1);
                   eps(2,2) = strain(2);
                   eps(3,3) = strain(3);
                   break;
         default:  throw featured_error( "undefined stress type in StrainVector2Tensor" );
      }
   }
   else
   {
      switch( st )
      {
         case gen: eps(3,1) = 0.5*strain(6);
                   eps(3,2) = 0.5*strain(5);
         case axi: eps(2,1) = 0.5*strain(4);
                   eps(1,1) = strain(1);
                   eps(2,2) = strain(2);
                   eps(3,3) = strain(3);
                   break;
         case psn:
         case pss: eps(2,1) = 0.5*strain(3);
                   eps(2,2) = strain(2);
         case uni: eps(1,1) = strain(1);
                   break;
         default:  throw featured_error( "undefined stress type in StrainVector2Tensor" );
      }
   }

   return eps;
}

void Tensor2CollectArray( const Tensor2Gen& T, CollectArray *C )
{
   for ( int i=1; i<4; i++ )
      for ( int j=1; j<4; j++ )
         C->add( T(i,j) );
}

void CollectArray2Tensor( const CollectArray &C, Tensor2Gen *T )
{
   for ( int i=1; i<4; i++ )
      for ( int j=1; j<4; j++ )
         (*T)(i,j) = C.extract();
}

