// Copyright (c) 1999, A.H. van den Boogaard

#ifndef ELASTIC_H
#define ELASTIC_H

#include "material.h"
#include "thermal.h"

/* advance declaration */
class Elastic_type;

class Elastic_Point:public FeaTure::Structural_Point
{
   protected:
      Elastic_type *parameters;
      FeaTure::Vector strain;
      FeaTure::Vector new_strain;

      FeaTure::Matrix get_xp_D_mat() const;

   public:
      Elastic_Point( Elastic_type *mat, FeaTure::StressType st );
      ~Elastic_Point();

      FeaTure::Matrix get_D_mat() const;
      FeaTure::Vector get_stress() const;
      FeaTure::Vector get_xp_stress() const;
      void set_strain_increment( const FeaTure::Vector& strain_increment );
      void finish_increment();

      void GetRestartData( FeaTure::CollectArray &data ) const;
      void PutRestartData( const FeaTure::CollectArray &data );
};

/* advance declaration */
class Elastic_type_xp;

class Elastic_Point_xp:public Elastic_Point
{
   private:
      FeaTure::Vector strain_bar;
      FeaTure::Vector prev_dstrain;

   public:
      Elastic_Point_xp( Elastic_type *mat, FeaTure::StressType st );
      ~Elastic_Point_xp();

      FeaTure::Matrix get_D_mat() const;
      FeaTure::Vector get_stress() const;
      FeaTure::Vector get_xp_stress() const;
      FeaTure::Matrix get_xp_D_mat() const;
      void finish_increment();
      void set_strain_increment( const FeaTure::Vector& strain_increment );
};

/* advance declaration */
class ThermoElastic_type;

class ThermoElastic_Point:public Elastic_Point, public LinearThermal_Point
{
   protected:
      double thermal_strain_old, Delta_eps_th;

   public:
      ThermoElastic_Point( ThermoElastic_type *mat, FeaTure::StressType st );
      ~ThermoElastic_Point();

      void set_strain_increment( const FeaTure::Vector& strain_increment );
      void set_potential_increment( double Tinc );
      void finish_increment();
      double get_expansion_coefficient() const;
      FeaTure::Vector get_xp_stress() const;
      FeaTure::Vector get_stress() const;

      void GetRestartData( FeaTure::CollectArray &data ) const;
      void PutRestartData( const FeaTure::CollectArray &data );
};

/* advance declaration */
class NonlinearElastic_type;

class NonlinearElastic_Point:public FeaTure::Structural_Point
{
   private:
      NonlinearElastic_type *parameters;
      FeaTure::Vector strain;
      FeaTure::Vector new_strain;

      FeaTure::Matrix get_xp_D_mat() const;

   public:
      NonlinearElastic_Point( NonlinearElastic_type *mat,
                              FeaTure::StressType st );
      ~NonlinearElastic_Point();

      FeaTure::Matrix get_D_mat() const;
      FeaTure::Vector get_stress() const;
      void set_strain_increment( const FeaTure::Vector& strain_increment );
      void finish_increment();

      void GetRestartData( FeaTure::CollectArray &data ) const;
      void PutRestartData( const FeaTure::CollectArray &data );
};

/* Material type definitions */

class Elastic_type: public FeaTure::Structural_type
{
   private:
      double Young, Poisson;

   public:
      Elastic_type( int matnr );
      ~Elastic_type();

      int set_Young( double young );
      double get_Young() const;
      int set_Poisson( double poisson );
      double get_Poisson() const;

#ifdef NOVARIABLERETURN
      FeaTure::Material_Point* new_Material_Point( FeaTure::StressType st );
#else
      Elastic_Point* new_Material_Point( FeaTure::StressType st );
#endif
      bool symmetric() const { return true; }
};

class Elastic_type_xp:public Elastic_type
{
   public:
      Elastic_type_xp( int matnr );
      ~Elastic_type_xp();

#ifdef NOVARIABLERETURN
      FeaTure::Material_Point* new_Material_Point( FeaTure::StressType st );
#else
      Elastic_Point_xp* new_Material_Point( FeaTure::StressType st );
#endif
      bool NeedExpandStrain( FeaTure::StressType /*st*/ ) const { return true; }
};

class ThermoElastic_type:public Elastic_type, public LinearThermal_type
{
   private:
      double alpha;

   public:
      ThermoElastic_type( int matnr );
      ~ThermoElastic_type();

#ifdef NOVARIABLERETURN
      FeaTure::Material_Point* new_Material_Point( FeaTure::StressType st );
#else
      ThermoElastic_Point* new_Material_Point( FeaTure::StressType st );
#endif
      int set_alpha( double al );
      double get_alpha() const { return alpha; }
      bool symmetric() const { return false; }
};

class NonlinearElastic_type:public FeaTure::Structural_type
{
   private:
      double Young;

   public:
      NonlinearElastic_type( int matnr );
      ~NonlinearElastic_type();

      int set_Young( double young );
      double get_Young() const;

#ifdef NOVARIABLERETURN
      FeaTure::Material_Point* new_Material_Point( FeaTure::StressType st );
#else
      NonlinearElastic_Point* new_Material_Point( FeaTure::StressType st );
#endif
      bool symmetric() const { return true; }
};

#endif /* ELASTIC_H */
