// Copyright (c) 1999,2003, A.H. van den Boogaard

#include "elements.h"
#include "constant.h"
#include "myexcept.h"
#include "elastic.h"

#include <iomanip>

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::CollectArray;
using FeaTure::Location;
using FeaTure::Direction;
using FeaTure::CartVector;
using FeaTure::Material_type;
using FeaTure::Material_Point;
using FeaTure::MatPntPtr;
using FeaTure::Structural_type;
using FeaTure::Structural_Point;
using FeaTure::Thermal_type;
using FeaTure::Thermal_Point;
using FeaTure::Mass_type;
using FeaTure::Geometry;
using FeaTure::Dof;
using FeaTure::StressType;
using FeaTure::IntLocs;
using FeaTure::PostProc;
using FeaTure::gen;
using FeaTure::uni;
using FeaTure::psn;
using FeaTure::pss;
using FeaTure::axi;
using FeaTure::get_nodal_dimension;

using std::cerr;
using std::istream;
using std::ostream;
using std::ios;
using std::setw;
using std::endl;

Truss::Truss( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom
              ):FiniteElement( nr, Nodes, 2, 2 )
{
   material_ptr = mat;
   geometry_ptr = geom;

   mat_point = dynamic_cast<Structural_Point *>(mat->new_Material_Point( uni ));
}

Truss::~Truss(){}

int Truss::check_data( const Material_type *mat, const Geometry *geom )
{
   int errors;
   errors = check_mgd_ptrs( mat, geom, 1, 1 );
   if ( geom != NULL && geom->get_Crosse() <= 0.0 )
   {
      cerr << "cross section <= 0; this is not allowed" << endl;
      errors++;
   }
   return errors;
}

bool Truss::NeedExpandStrain() const
{
   return dynamic_cast<Structural_type*>(material_ptr)->NeedExpandStrain( uni );
}

FiniteElement::LocalDof* Truss::GetLocalDofs() const
{
   Location coordi1, coordi2;
   LocalDof* ldofs;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();

   Direction el_dir( coordi2 - coordi1 );

   ldofs = new LocalDof[nr_of_local_dofs];

   int idof = 0;
   for ( int i=1; i<=nr_of_element_nodes; i++ )
   {
      ldofs[idof].dir = el_dir;
      ldofs[idof].phys_type = Dof::u;
      ldofs[idof].loc_node_nr = i;
      idof++;
   }
   if ( idof != nr_of_local_dofs )
      throw featured_error( "error in Truss::GetLocalDofs()" );

   return ldofs;
}

Matrix Truss::GetLocalStiff( Vector& stress_correction ) const
{
   double stiff, E, length;
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   length = ( coordi2 - coordi1 ).norm();

   E = mat_point->get_D_mat()(1,1);
   stiff = E*geometry_ptr->get_Crosse()/length;
   Matrix Stifmat(2,2);
   Stifmat(1,1) = stiff;
   Stifmat(1,2) =-stiff;
   Stifmat(2,1) =-stiff;
   Stifmat(2,2) = stiff;

   stress_correction.resize(2);
   stress_correction.clear();

   return Stifmat;
}

Matrix Truss::GetLocalMass( bool lumped ) const
{
   double A, L, rho;
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   L = ( coordi2 - coordi1 ).norm();

   A = geometry_ptr->get_Crosse();
   rho = material_ptr->get_density();

   double a(L/2);

   Matrix M(2,2);

   if ( lumped )
   {
      double ml( A*rho*a );

      M[0][0] = ml;
      M[0][1] = 0;
      M[1][0] = 0;
      M[1][1] = ml;
   }
   else
   {
      double ml( A*rho*a/3.0 );
   
      M[0][0] = ml*2;
      M[0][1] = ml;
      M[1][0] = ml;
      M[1][1] = ml*2;
   }

   return M;
}

Vector Truss::GetLocalLHS() const
{
   double Force;

   Vector sigma(mat_point->get_stress());
   Force = sigma(1)*geometry_ptr->get_Crosse();

   Vector LHS(2);
   LHS(1) = -Force;
   LHS(2) =  Force;

   return LHS;
}

void Truss::iteration_update()
{
   double   length;
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   length = ( coordi2 - coordi1 ).norm();

   Vector dU( get_local_increment_vector() );

   Vector dstrain(1);
   dstrain(1) = ( dU(2) - dU(1) ) / length;

   mat_point->set_strain_increment( dstrain );
}

void Truss::finish_increment()
{
   mat_point->finish_increment();
}

void Truss::WriteElementRestartData( ostream& restart ) const
{
   CollectArray data;
   mat_point->GetRestartData( data );
   restart << data;
}

void Truss::ReadElementRestartData( istream& restart )
{
   CollectArray data;
   restart >> data;
   mat_point->PutRestartData( data );
}

void Truss::PrintStressHeader( ostream& out ) const
{
   out << endl << "             Sxx" << endl;
}

void Truss::PrintXpStressHeader( ostream& out ) const
{
   out << endl << "             Sxx         Syy         Szz" << endl;
}

void Truss::PrintStress( ostream& out ) const
{
   out << "element " << elm_nr << endl;

   out.setf( ios::scientific, ios::floatfield );
   out.precision(3);
   out << "        " << setw(11);
   (mat_point->get_stress()).print_row( out );
   out << endl << endl;
}

void Truss::PrintXpStress( ostream& out ) const
{
   out << "element " << elm_nr << endl;

   out.setf( ios::scientific, ios::floatfield );
   out.precision(3);
   out << "        " << setw(11);
   (mat_point->get_xp_stress()).print_row(out);
   out << endl << endl;
}

#ifdef FT_GUI
void Truss::Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const
{
   double x1, y1, x2, y2, z1, z2;
   wxCoord x1i, y1i, x2i, y2i;

   Nodes[0]->get_location( x1, y1, z1 );
   Nodes[1]->get_location( x2, y2, z2 );
   x1i=map.XtoScreen(x1);
   x2i=map.XtoScreen(x2);
   y1i=map.YtoScreen(y1);
   y2i=map.YtoScreen(y2);

   dc.DrawLine( x1i, y1i, x2i, y2i );
   if ( numbered )
   {
      char s[20];
      sprintf(s,"%d",elm_nr);
      dc.DrawText(s,(x1i+x2i)/2,(y1i+y2i)/2);
   }
}

void Truss::DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered, float magnif ) const
{
   double x1, y1, x2, y2, z1, z2;
   wxCoord x1i, y1i, x2i, y2i;

   Nodes[0]->draw_location( x1, y1, z1, magnif );
   Nodes[1]->draw_location( x2, y2, z2, magnif );
   x1i=map.XtoScreen(x1);
   x2i=map.XtoScreen(x2);
   y1i=map.YtoScreen(y1);
   y2i=map.YtoScreen(y2);

   dc.DrawLine( x1i, y1i, x2i, y2i );
   if ( numbered )
   {
      char s[20];
      sprintf(s,"%d",elm_nr);
      dc.DrawText(s,(x1i+x2i)/2,(y1i+y2i)/2);
   }
}
#endif // FT_GUI

LinBeam2d::LinBeam2d( int nr, Node_ptr *Nodes, Material_type *mat,
                      Geometry *geom) : FiniteElement( nr, Nodes, 2, 6 )
{
   material_ptr = mat;
   geometry_ptr = geom;

   LHS_begin.resize(6);
   LHS_begin.clear();
   LHS_end.resize(6);
   LHS_end.clear();
}

LinBeam2d::~LinBeam2d(){}

int LinBeam2d::check_data( const Material_type *mat, const Geometry *geom )
{
   int errors;
   errors = check_mgd_ptrs( mat, geom, 1, 1 );
   if ( geom != NULL && (geom->get_Crosse() <= 0.0 || geom->get_Iz() <= 0.0) )
   {
      cerr << "cross section or moment of inertia <= 0; this is not allowed"
           << endl;
      errors++;
   }
   return errors;
}

FiniteElement::LocalDof* LinBeam2d::GetLocalDofs() const
{
   Location coordi1, coordi2;
   LocalDof* ldofs;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();

   Direction el_dir( coordi2 - coordi1 );

   if ( std::fabs(el_dir[2])>1.e-15 )
      throw featured_error(
                "2D beam element should be parallel to the x-y plane" );
   Direction dir3;
   dir3.set_dir( 0, 0, 1 );
   CartVector yaxis( FeaTure::outer(dir3,el_dir) );
   Direction dir2( yaxis );

   ldofs = new LocalDof[nr_of_local_dofs];

   int idof = 0;
   for ( int i=1; i<=nr_of_element_nodes; i++ )
   {
      ldofs[idof].dir = el_dir;
      ldofs[idof].phys_type = Dof::u;
      ldofs[idof].loc_node_nr = i;
      idof++;
      ldofs[idof].dir = dir2;
      ldofs[idof].phys_type = Dof::u;
      ldofs[idof].loc_node_nr = i;
      idof++;
      ldofs[idof].dir = dir3;
      ldofs[idof].phys_type = Dof::r;
      ldofs[idof].loc_node_nr = i;
      idof++;
   }
   if ( idof != nr_of_local_dofs )
      throw featured_error( "error in LinBeam2d::GetLocalDofs()" );

   return ldofs;
}

Matrix LinBeam2d::GetLocalStiff( Vector& /*stress_correction*/ ) const
{
   double astiff, bstiff, E, length;
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   length = ( coordi2 - coordi1 ).norm();

   E = dynamic_cast<Elastic_type*>(material_ptr)->get_Young();
   astiff = E*geometry_ptr->get_Crosse()/length;
   bstiff = E*geometry_ptr->get_Iz()/(length*length*length);
   Matrix Stifmat(6,6,0.0);

   // truss part
   Stifmat(1,1) =  astiff;
   Stifmat(1,4) = -astiff;
   Stifmat(4,1) = -astiff;
   Stifmat(4,4) =  astiff;

   Stifmat(2,2) =  12*bstiff;
   Stifmat(2,5) = -12*bstiff;
   Stifmat(5,2) = -12*bstiff;
   Stifmat(5,5) =  12*bstiff;

   bstiff *= length;

   Stifmat(2,3) =  6*bstiff;
   Stifmat(2,6) =  6*bstiff;
   Stifmat(3,2) =  6*bstiff;
   Stifmat(3,5) = -6*bstiff;
   Stifmat(5,3) = -6*bstiff;
   Stifmat(5,6) = -6*bstiff;
   Stifmat(6,2) =  6*bstiff;
   Stifmat(6,5) = -6*bstiff;

   bstiff *= length;

   Stifmat(3,3) =  4*bstiff;
   Stifmat(3,6) =  2*bstiff;
   Stifmat(6,3) =  2*bstiff;
   Stifmat(6,6) =  4*bstiff;

   return Stifmat;
}

Vector LinBeam2d::GetLocalLHS() const
{
   return LHS_end;
}

Matrix LinBeam2d::GetLocalMass( bool lumped ) const
{
   double A, L, rho;
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   L = ( coordi2 - coordi1 ).norm();

   A = geometry_ptr->get_Crosse();
   rho = material_ptr->get_density();

   double a(L/2);

   Matrix M(6,6,0.0);

   if ( lumped )
   {
      // proportional scaling
      // (Petyt, Introduction to finite element vibration analysis, p. 442)
      double ml( A*rho*a );

      M[0][0] = ml;
      M[1][1] = ml;
      M[2][2] = ml*4*a*a/39;
      M[3][3] = ml;
      M[4][4] = ml;
      M[5][5] = M[2][2];
   }
   else
   {
      // consistent mass matrix
      // (Petyt, Introduction to finite element vibration analysis, p. 89)
      double ml( A*rho*a/3.0 );
      double mt( A*rho*a/105.0 );
   
      M[0][0] = ml*2;
      M[0][3] = ml;
      M[3][0] = ml;
      M[3][3] = ml*2;
      M[1][1] = mt*78;
      M[1][2] = mt*22*a;
      M[1][4] = mt*27;
      M[1][5] = -mt*13*a;
      M[2][1] = M[1][2];
      M[2][2] = mt*8*a*a;
      M[2][4] = mt*13*a;
      M[2][5] = -mt*6*a*a;
      M[4][1] = M[1][4];
      M[4][2] = M[2][4];
      M[4][4] = mt*78;
      M[4][5] = -mt*22*a;
      M[5][1] = M[1][5];
      M[5][2] = M[2][5];
      M[5][4] = M[4][5];
      M[5][5] = mt*8*a*a;
   }

   return M;
}

void LinBeam2d::iteration_update()
{
   Vector dU( get_local_increment_vector() );

   Vector dummy(1,0);
   Matrix K( GetLocalStiff( dummy ) );

   LHS_end = LHS_begin + K*dU;
}

void LinBeam2d::finish_increment()
{
   LHS_begin = LHS_end;
}

void LinBeam2d::WriteElementRestartData( ostream& restart ) const
{
   CollectArray data;
   data.add(LHS_begin);
   restart << data;
}

void LinBeam2d::ReadElementRestartData( istream& restart )
{
   CollectArray data;
   restart >> data;
   data.extract( LHS_begin );
   LHS_end = LHS_begin;
}

void LinBeam2d::PrintStressHeader( ostream& out ) const
{
   PrintXpStressHeader( out );
}

void LinBeam2d::PrintXpStressHeader( ostream& out ) const
{
   out << endl << "                 Nx          Sy          Mz " << endl;
}

void LinBeam2d::PrintStress( ostream& out ) const
{
   PrintXpStress( out );
}

void LinBeam2d::PrintXpStress( ostream& out ) const
{
   Vector FG;
   out << "element " << elm_nr << endl;

   out.setf( ios::scientific, ios::floatfield );
   out.precision(3);

   FG = get_gen_forces( 0.0 );
   out << " node " << setw(4) << Nodes[0]->get_nr() << "  " << setw(11);
   FG.print_row(out);
   out << endl;
   FG = get_gen_forces( 1.0 );
   out << "      " << setw(4) << Nodes[1]->get_nr() << "  " << setw(11);
   FG.print_row(out);
   out << endl << endl;
}

FeaTure::Matrix LinBeam2d::get_N_mat( double L2 ) const
{
   // Interpolation matrix such that  u=N*dof gives a vector
   // {u} with the displacement in local x- and y-direction

   double   length;
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   length = ( coordi2 - coordi1 ).norm();

   Matrix N_mat( 2, 6, 0.0 );
   double L22( L2*L2 );
   double L23( L22*L2 );

   N_mat(1,1) = 1-L2;
   N_mat(2,2) = 1-3*L22+2*L23;
   N_mat(2,3) = length*(L2-2*L22+L23);
   N_mat(1,4) = L2;
   N_mat(2,5) = 3*L22-2*L23;
   N_mat(2,6) = length*(-L22+L23);

   return N_mat;
}

FeaTure::Matrix LinBeam2d::get_B_mat( double L2 ) const
{
   // B matrix such that  u=B*dof gives a vector
   // {u} with du/dx,  -d2v/dx2
   // conjugated to N, M (normal force, moment

   double   invlength, invlength2;
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   invlength = 1.0/(( coordi2 - coordi1 ).norm());
   invlength2 = invlength*invlength;

   Matrix B_mat( 2, 6, 0.0 );

   B_mat(1,1) = -invlength;
   B_mat(2,2) = (6-12*L2)*invlength2;
   B_mat(2,3) = (4- 6*L2)*invlength;
   B_mat(1,4) =  invlength;
   B_mat(2,5) = (-6+12*L2)*invlength2;
   B_mat(2,6) = ( 2-6*L2)*invlength;

   return B_mat;
}

FeaTure::Vector LinBeam2d::get_gen_forces( double L2 ) const
{
   // Get generalised forces N, S, M (normal force, shear force, moment)
   // with respect to local coordinate system

   double   invlength, invlength2, invlength3;
   Location coordi1, coordi2;

   double EA, EI, E;
   E = dynamic_cast<Elastic_type*>(material_ptr)->get_Young();
   EA = E*geometry_ptr->get_Crosse();
   EI = E*geometry_ptr->get_Iz();

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();
   invlength = 1.0/(( coordi2 - coordi1 ).norm());
   invlength2 = invlength*invlength;
   invlength3 = invlength2*invlength;

   Vector U( get_local_total_vector() );

   Vector Fgen(3);
   Fgen(1) = EA*invlength*(U(4)-U(1));
   Fgen(2) = EI*(-12*invlength3*(U(2)-U(5))-6*invlength2*(U(3)+U(6)));
   Fgen(3) = EI*(((-6+12*L2)*U(2)+(6-12*L2)*U(5))*invlength2 +
                 ((-4+6*L2)*U(3)+(-2+6*L2)*U(6))*invlength);

   return Fgen;
}

#ifdef FT_GUI
void LinBeam2d::Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const
{
   double x1, y1, x2, y2, z1, z2;
   int    x1i, y1i, x2i, y2i;

   Nodes[0]->get_location( x1, y1, z1 );
   Nodes[1]->get_location( x2, y2, z2 );
   x1i=int(map.XtoScreen(x1));
   x2i=int(map.XtoScreen(x2));
   y1i=int(map.YtoScreen(y1));
   y2i=int(map.YtoScreen(y2));

   dc.DrawLine( x1i, y1i, x2i, y2i );
   if ( numbered )
   {
      char s[20];
      sprintf(s,"%d",elm_nr);
      dc.DrawText(s,(x1i+x2i)/2,(y1i+y2i)/2);
   }
}

void LinBeam2d::DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered, float magnif ) const
{
   Location coordi1, coordi2;

   coordi1 = Nodes[0]->get_location();
   coordi2 = Nodes[1]->get_location();

   Direction el_dir( coordi2 - coordi1 );
   double c( el_dir(1) );  // cosine
   double s( el_dir(2) );  // sine

   double x1, y1, z1;
   Nodes[0]->draw_location( x1, y1, z1, magnif );

   wxCoord x1i, y1i, x2i, y2i;

   x1i=map.XtoScreen(x1);
   y1i=map.YtoScreen(y1);

   Vector dof_total;

   if ( nr_of_system_dofs==0 )
   {  // if only read, not run:
      dof_total.resize(6);
      dof_total.clear();
   }
   else
      dof_total = get_local_total_vector();

   for ( int i=1; i<9; i++ )
   {
      double L2( i*0.125 );
      Matrix N( get_N_mat(L2) );
      Vector u( (N*dof_total)*magnif );
      Location coordi( (1-L2)*coordi1 + L2*coordi2 );
      coordi(1) += c*u(1) - s*u(2);
      coordi(2) += s*u(1) + c*u(2);

      x2i=map.XtoScreen(coordi(1));
      y2i=map.YtoScreen(coordi(2));

      dc.DrawLine( x1i, y1i, x2i, y2i );

      if ( numbered && i==4 )
      {
         char s[20];
         sprintf(s,"%d",elm_nr);
         dc.DrawText(s,x2i,y2i);
      }

      x1i = x2i;
      y1i = y2i;
   }
}
#endif // FT_GUI


Mass::Mass( int nr, Node_ptr *Nodes, Material_type *mat, Geometry */*geom*/,
            int nr_dofs ):FiniteElement( nr, Nodes, 1, nr_dofs )
{
   material_ptr = dynamic_cast<Mass_type*>(mat);
}

int Mass::check_data( const Material_type *mat, const Geometry *geom )
{
   int errors;
   errors = check_mgd_ptrs( mat, geom, 1, 0 );

   if ( mat != NULL && dynamic_cast<const Mass_type*>(mat)->get_mass() <= 0.0 )
   {
      cerr << "mass <= 0; this is not allowed" << endl;
      errors++;
   }

   if ( geom != NULL )
   {
      cerr << "geometry definition is not allowed for mass element" << endl;
      errors++;
   }

   return errors;
}

Matrix Mass::GetLocalStiff( Vector& stress_correction ) const
{
   Matrix Stifmat(nr_of_local_dofs,nr_of_local_dofs,0.0);
   stress_correction.resize(nr_of_local_dofs);
   stress_correction.clear();

   return Stifmat;
}

Vector Mass::GetLocalLHS() const
{
   Vector LHS(nr_of_local_dofs, 0);
   return LHS;
}

Matrix Mass::GetLocalMass( bool /*lumped*/ ) const
{
   Matrix M(nr_of_local_dofs,nr_of_local_dofs,0.0);
   double mass = material_ptr->get_mass();
   for ( int i=0; i<nr_of_local_dofs; i++ ) M[i][i] = mass;

   return M;
}

#ifdef FT_GUI
void Mass::Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const
{
   double x, y, z;
   wxCoord xi, yi;

   Nodes[0]->get_location( x, y, z );
   xi=map.XtoScreen(x);
   yi=map.YtoScreen(y);

   dc.DrawEllipse( xi-2, yi-2, 4, 4 );
   if ( numbered )
   {
      char s[20];
      wxCoord h;
      sprintf(s,"%d",elm_nr);
      h = dc.GetCharHeight();
      dc.DrawText( s, xi, yi-h );
   }
}

void Mass::DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered, float magnif ) const
{
   double x, y, z;
   wxCoord xi, yi;

   Nodes[0]->draw_location( x, y, z, magnif );
   xi=map.XtoScreen(x);
   yi=map.YtoScreen(y);

   dc.DrawEllipse( xi-2, yi-2, 4, 4 );
   if ( numbered )
   {
      char s[20];
      wxCoord h;
      sprintf(s,"%d",elm_nr);
      h = dc.GetCharHeight();
      dc.DrawText( s, xi, yi-h );
   }
}
#endif // FT_GUI

Mass2D::Mass2D( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom )
        :Mass( nr, Nodes, mat, geom, 2 ){}

FiniteElement::LocalDof* Mass2D::GetLocalDofs() const
{
   Direction el_dir;
   LocalDof* ldofs;

   ldofs = new LocalDof[nr_of_local_dofs];

   el_dir.set_dir( 1, 0, 0 );
   ldofs[0].dir = el_dir;
   ldofs[0].phys_type = Dof::u;
   ldofs[0].loc_node_nr = 1;

   el_dir.set_dir( 0, 1, 0 );
   ldofs[1].dir = el_dir;
   ldofs[1].phys_type = Dof::u;
   ldofs[1].loc_node_nr = 1;

   return ldofs;
}

Mass3D::Mass3D( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom )
        :Mass( nr, Nodes, mat, geom, 3 ){}

FiniteElement::LocalDof* Mass3D::GetLocalDofs() const
{
   Direction el_dir;
   LocalDof* ldofs;

   ldofs = new LocalDof[nr_of_local_dofs];

   el_dir.set_dir( 1, 0, 0 );
   ldofs[0].dir = el_dir;
   ldofs[0].phys_type = Dof::u;
   ldofs[0].loc_node_nr = 1;

   el_dir.set_dir( 0, 1, 0 );
   ldofs[1].dir = el_dir;
   ldofs[1].phys_type = Dof::u;
   ldofs[1].loc_node_nr = 1;

   el_dir.set_dir( 0, 0, 1 );
   ldofs[2].dir = el_dir;
   ldofs[2].phys_type = Dof::u;
   ldofs[2].loc_node_nr = 1;

   return ldofs;
}

NumIntElement::NumIntElement():FiniteElement()
{
   material_ptr = NULL;
   geometry_ptr = NULL;
   mat_point    = NULL;
}

NumIntElement::NumIntElement( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                              Material_type *mat, Geometry *geom, int nip )
              :FiniteElement( nr, Nodes, nr_nodes, nr_dofs )
{
   material_ptr = mat;
   geometry_ptr = geom;
   shape        = 0;
#ifdef FT_GUI
   draw_shape   = 0;
#endif

   mat_point = new MatPntPtr[nip];

   for ( int ip=0; ip<nip; ip++ )
      mat_point[ip] = mat->new_Material_Point( gen );
}

NumIntElement::~NumIntElement()
{
   delete [] mat_point;
   delete shape;
#ifdef FT_GUI
   delete draw_shape;
#endif
}

void NumIntElement::finish_increment()
{
   for ( int ip=0; ip<get_nr_int_points(); ip++ )
      mat_point[ip]->finish_increment();
}

void NumIntElement::WriteElementRestartData( ostream& restart ) const
{
   CollectArray data;

   for ( int ip=0; ip<get_nr_int_points(); ip++ )
      mat_point[ip]->GetRestartData( data );

   restart << data;
}

void NumIntElement::ReadElementRestartData( istream& restart )
{
   CollectArray data;
   restart >> data;

   for ( int ip=0; ip<get_nr_int_points(); ip++ )
      mat_point[ip]->PutRestartData( data );
}

void NumIntElement::GetCoordinateMatrix( Matrix& coor, int ndim ) const
{
   double z;

   coor.resize( nr_of_element_nodes, ndim );
   if ( ndim == 2 )
      for ( int i=1; i<=nr_of_element_nodes; i++ )
         Nodes[i-1]->get_location(coor(i,1), coor(i,2), z);
   else if ( ndim == 3 )
      for ( int i=1; i<=nr_of_element_nodes; i++ )
         Nodes[i-1]->get_location(coor(i,1), coor(i,2), coor(i,3));
   else
      throw featured_error( "unknown nr. of nodal dimensions in getCoordinateMatrix()" );
}

void NumIntElement::GetNewCoordinateMatrix( Matrix& coor, int ndim ) const
{
   double z;

   coor.resize( nr_of_element_nodes, ndim );
   if ( ndim == 2 )
      for ( int i=1; i<=nr_of_element_nodes; i++ )
         Nodes[i-1]->get_current_location( coor(i,1),coor(i,2),z );
   else if ( ndim == 3 )
      for ( int i=1; i<=nr_of_element_nodes; i++ )
         Nodes[i-1]->get_current_location( coor(i,1), coor(i,2), coor(i,3) );
   else
      throw featured_error( "unknown nr. of nodal dimensions in getCoordinateMatrix()" );
}

#ifdef FT_GUI
void NumIntElement::Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const
{
   int         npoints;
   double      x, y, z;
   wxCoord     xsum, ysum;
   std::vector<int> points;
   wxPoint     *pt;

   if ( draw_shape == 0 ) return;

   draw_shape->get_draw_points( points );
   npoints = points.size();

   pt = new wxPoint[npoints];

   xsum = ysum = 0;
   for ( int i=0; i<npoints; i++ )
   {
      Nodes[points[i]]->get_location(x,y,z);
      pt[i] = map.toScreen(x,y);
      if ( numbered )
      {
         xsum += pt[i].x;
         ysum += pt[i].y;
      }
   }

   dc.DrawPolygon(npoints,pt);
   delete [] pt;

   if ( numbered )
   {
      wxString s;
      wxCoord w, h;
      s.Printf("%d",elm_nr);
      dc.GetTextExtent(s,&w,&h);
      dc.DrawText( s, xsum/npoints-w/2, ysum/npoints-h/2 );
   }
}

void NumIntElement::DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered, float magnif ) const
{
   int         npoints;
   double      x, y, z;
   wxCoord     xsum, ysum;
   std::vector<int> points;
   wxPoint     *pt;

   if ( draw_shape == 0 ) return;

   draw_shape->get_draw_points( points );
   npoints = points.size();

   pt = new wxPoint[npoints];

   xsum = ysum = 0;
   for ( int i=0; i<npoints; i++ )
   {
      Nodes[points[i]]->draw_location(x,y,z, magnif);
      pt[i] = map.toScreen(x,y);
      if ( numbered )
      {
         xsum += pt[i].x;
         ysum += pt[i].y;
      }
   }

   dc.DrawPolygon(npoints,pt);
   delete [] pt;

   if ( numbered )
   {
      wxString s;
      wxCoord w, h;
      s.Printf("%d",elm_nr);
      dc.GetTextExtent(s,&w,&h);
      dc.DrawText( s, xsum/npoints-w/2, ysum/npoints-h/2 );
   }
}
#endif // FT_GUI

NumIntStructElement::NumIntStructElement( int nr, Node_ptr *ElNodes, int nr_nodes,
                              int struct_dofs, int nr_dofs,
                              Material_type *mat, Geometry *geom,
                              StressType st, int nip ):NumIntElement()
{
   elm_nr = nr;
   Nodes = ElNodes;
   nr_of_element_nodes = nr_nodes;
   nr_of_local_dofs = nr_dofs;
   material_ptr = mat;
   geometry_ptr = geom;
   stype        = st;
   nr_struct_dofs = struct_dofs;

   mat_point = new MatPntPtr[nip];

   for ( int ip=0; ip<nip; ip++ )
      mat_point[ip] = mat->new_Material_Point( st );
}

NumIntStructElement::~NumIntStructElement(){}

Matrix NumIntStructElement::get_B_mat( const Vector &natcor, const Matrix &coor, double* DetJ ) const
{
   int    ndim, nstr, ndof, i;

   ndim = get_nodal_dimension( stype );
   nstr = StressSize( stype );
   ndof = ndim*nr_of_element_nodes;

   Matrix dNdXi = shape->getPartialDerivatives( natcor );

   Matrix J = dNdXi*coor;
   *DetJ = J.determinant();

   Matrix dNdX = J.inv()*dNdXi;

   Matrix B(nstr,ndof,0.0);

   if ( stype == psn || stype == pss )
   {
      if ( stype==pss ) *DetJ *= geometry_ptr->get_Thick();

      for ( i=1; i<=nr_of_element_nodes; i++ )
      {
         B(1,2*i-1) = dNdX(1,i);
         B(2,2*i)   = dNdX(2,i);
         B(3,2*i-1) = dNdX(2,i);
         B(3,2*i)   = dNdX(1,i);
      }
   }
   else if ( stype == axi )
   {
      Matrix N = shape->get_N_vec( natcor );
      Matrix x = N*coor;

      assert( x(1,1) >= 1.e-20 );  /* radius is always > 0, therefore abs() is not necessary */
      double rinv = 1/x(1,1);
      *DetJ *= PI2*x(1,1);

      for ( i=1; i<=nr_of_element_nodes; i++ )
      {
         B(1,2*i-1) = dNdX(1,i);
         B(2,2*i)   = dNdX(2,i);
         B(3,2*i-1) = rinv*N(1,i);
         B(4,2*i-1) = dNdX(2,i);
         B(4,2*i)   = dNdX(1,i);
      }
   }
   else if ( stype == gen )
   {
      for ( i=1; i<=nr_of_element_nodes; i++ )
      {
         B(1,3*i-2) = dNdX(1,i);
         B(2,3*i-1) = dNdX(2,i);
         B(3,3*i  ) = dNdX(3,i);
         B(4,3*i-2) = dNdX(2,i);
         B(4,3*i-1) = dNdX(1,i);
         B(5,3*i-1) = dNdX(3,i);
         B(5,3*i  ) = dNdX(2,i);
         B(6,3*i-2) = dNdX(3,i);
         B(6,3*i  ) = dNdX(1,i);
      }
   }
   else
      throw featured_error( "wrong stress type in NumIntStructElement::get_B_mat()" );

   return B;
}

Matrix NumIntStructElement::get_N_mat( const Vector &natcor ) const
{
   int ndim, ndof;

   ndim = get_nodal_dimension( stype );
   ndof = ndim*nr_of_element_nodes;

   Matrix N_mat( ndim, ndof, 0.0 );
   Matrix N_vec( shape->get_N_vec( natcor ) );

   int idof=0;
   for ( int node=0; node<nr_of_element_nodes; node++ )
   {
      for ( int idim=0; idim<ndim; idim++ )
      {
         N_mat[idim][idof] = N_vec[0][node];
         idof++;
      }
   }

   return N_mat;
}

bool NumIntStructElement::NeedExpandStrain() const
{
   return dynamic_cast<Structural_type*>(material_ptr)->NeedExpandStrain( stype );
}

FiniteElement::LocalDof* NumIntStructElement::GetLocalDofs() const
{
   Direction el_dir[3];
   LocalDof* ldofs;

   int ndim(get_nodal_dimension( stype ));

   el_dir[0].set_dir( 1,0,0 );
   el_dir[1].set_dir( 0,1,0 );
   el_dir[2].set_dir( 0,0,1 );

   ldofs = new LocalDof[nr_struct_dofs];

   int idof = 0;
   for ( int node=1; node<=nr_of_element_nodes; node++ )
   {
      for ( int ndof=0; ndof<ndim; ndof++ )
      {
         ldofs[idof].dir = el_dir[ndof];
         ldofs[idof].phys_type = Dof::u;
         ldofs[idof].loc_node_nr = node;
         idof++;
      }
   }

   if ( idof != nr_struct_dofs )
      throw featured_error( "error in NumInt2D::GetLocalDofs()" );

   return ldofs;
}

Matrix NumIntStructElement::GetLocalStiff( Vector& stress_correction_vector ) const
{
   int    ip, ndof;
   double dV, DetJ;
   Matrix D;

   bool expand = NeedExpandStrain();

   ndof = nr_struct_dofs;

   Matrix Stifmat(ndof,ndof,0.0);

   if ( expand )
   {
      stress_correction_vector.resize(ndof);
      stress_correction_vector.clear();
   }

   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   IntLocs intlocs( get_int_locs() );

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      Matrix B( get_B_mat( intlocs.get_coor(ip), coor_mat, &DetJ ) );
      dV = DetJ*intlocs.get_weight( ip );

      if ( expand )
      {
         Vector stress_correction;
         dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_D_and_stress_correction( D, stress_correction );
         stress_correction_vector -= ~B*(stress_correction*dV);
      }
      else D = dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_D_mat();

//      Stifmat += ~B*D*B *dV;
      Add_sAtBA( Stifmat, 1.0, B, dV*D );
   }

   return Stifmat;
}

Vector NumIntStructElement::GetLocalLHS() const
{
   int    ip, ndof;
   double dV, DetJ;
   Matrix B;
   Vector sigma;

   ndof = nr_struct_dofs;

   Vector LHS(ndof,0);

   IntLocs intlocs( get_int_locs() );

   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      B = get_B_mat( intlocs.get_coor(ip), coor_mat, &DetJ );
      dV = DetJ*intlocs.get_weight( ip );
      sigma = dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_stress();
      LHS += ~B*(sigma *dV);
   }

   return LHS;
}

Matrix NumIntStructElement::GetLocalMass( bool lumped ) const
{
   int    ip, ndof;
   double dV, DetJ, rho;

   ndof = nr_struct_dofs;

   IntLocs intlocs( get_int_locs() );

   rho = material_ptr->get_density();
   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   Matrix M(nr_of_local_dofs,nr_of_local_dofs,0.0);
   Matrix Nv;

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      Vector natcor( intlocs.get_coor(ip) );

      Matrix dNdXi = shape->getPartialDerivatives( natcor );
      Matrix J = dNdXi*coor_mat;
      DetJ = J.determinant();

      if ( stype==pss ) DetJ *= geometry_ptr->get_Thick();
      else if ( stype == axi )
      {
         Matrix N = shape->get_N_vec( natcor );
         Matrix x = N*coor_mat;
         DetJ *= PI2*x(1,1);
      }

      Nv = get_N_mat( natcor );

      dV = DetJ*intlocs.get_weight( ip );
      M += ~Nv*Nv *(rho*dV);
   }

   if ( lumped )
   {
      for ( int row=0; row<ndof; row++ )
      {
         double sum=0;
         for ( int col=0; col<ndof; col++ ) sum += M[row][col];
         for ( int col=0; col<ndof; col++ ) M[row][col] = row==col?sum:0.0;
      }
   }

   return M;
}

void NumIntStructElement::iteration_update()
{
   IntLocs intlocs( get_int_locs() );

   Vector dU( get_local_increment_vector() );

   for ( int ip=1; ip<=get_nr_int_points(); ip++ )
      iteration_update_ip_struct( intlocs.get_coor(ip), dU,
                            dynamic_cast<Structural_Point*>(mat_point[ip-1]) );
}

void NumIntStructElement::iteration_update_ip_struct( const Vector& natcor,
                                 const Vector& dU, Structural_Point *matpoint )
{
   double DetJ;
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   Matrix B( get_B_mat( natcor, coor_mat, &DetJ ) );
   Vector dstrain = B*dU;

   matpoint->set_strain_increment( dstrain );
}

Vector NumIntStructElement::GetLocalPresVec( int surface, double magnitude ) const
{
   Vector Pload( nr_struct_dofs, 0.0 );

   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;
   GetNewCoordinateMatrix( coor_mat, ndim );

   Matrix Ns;

   IntLocs intlocs = shape->get_int_locs_face( surface );
   int nr_int_pt = intlocs.get_nr_intpt();
   Matrix dVdA = shape->dVdA( surface );

   for ( int ip=1; ip<=nr_int_pt; ip++ )
   {
      Matrix dNdXi = shape->getPartialDerivatives( intlocs.get_coor(ip) );
      Matrix J = dNdXi*coor_mat;
      // Matrix dXdS = (~J)*dVdA;
      Matrix dXdS = tmult(J,dVdA);

      Vector normal(ndim);
      if ( ndim == 2 )
      {
         normal(1) =  dXdS(2,1);
         normal(2) = -dXdS(1,1);
      }
      else if ( ndim == 3 )
      {
         normal(1) =  dXdS(2,1)*dXdS(3,2)-dXdS(2,2)*dXdS(3,1);
         normal(2) = -dXdS(1,1)*dXdS(3,2)+dXdS(1,2)*dXdS(3,1);
         normal(3) =  dXdS(1,1)*dXdS(2,2)-dXdS(1,2)*dXdS(2,1);
      }
//      cout << "normal " << normal << endl;

      double dV = intlocs.get_weight( ip );
      Ns = shape->get_N_vec( intlocs.get_coor(ip) );

      if ( stype==pss ) dV *= geometry_ptr->get_Thick();
      else if ( stype == axi )
      {
         Matrix x = Ns*coor_mat;
         dV *= PI2*x(1,1);
      }

      int idof = 0;
      for ( int inode=0; inode<nr_of_element_nodes; inode++ )
         for ( int kdof=0; kdof<ndim; kdof++ )
         {
            /* normal is outward, pressure is inward */
            Pload[idof] -= Ns[0][inode]*normal[kdof]*magnitude*dV;
            idof++;
         }
   }

   return Pload;
}

void NumIntStructElement::PrintStress( ostream& out ) const
{
   int    ip;

   out << "element " << elm_nr << endl;

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      out << setw(7) << ip;

      out.setf( ios::scientific, ios::floatfield );
      out.precision(3);
      out << " " << setw(11);
      (dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_stress()).print_row(out);
      out << endl;
   }
   out << endl;
}

void NumIntStructElement::PrintXpStress( ostream& out ) const
{
   int    ip;

   out << "element " << elm_nr << endl;

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      out << setw(7) << ip;

      out.setf( ios::scientific, ios::floatfield );
      out.precision(3);
      out << " " << setw(11);
      (dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_xp_stress()).print_row(out);
      out << endl;
   }
   out << endl;
}

void NumIntStructElement::PrintPostHeader( ostream &out, int stepnr ) const
{
   switch ( GetPostItem() )
   {
      case PostProc::coordinates:
                         out << "Gauss_points   0  " << setw(7) << get_nr_int_points()
                             << " 0   0   0" << endl;
                         break;
      case PostProc::stresses:
                         out << "Stresses   1   " << setw(7) << stepnr << " 3   1   0" << endl;
                         break;
      case PostProc::eps_plast_eq:
                         out << "Eq_Pl_strain   1   " << setw(7) << stepnr << " 1   1   0" << endl;
                         break;
   }

   total_ip_nr = 0;
}

void NumIntStructElement::PrintPostFile( ostream& out ) const
{
   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;

   IntLocs intlocs( get_int_locs() );

   if ( GetPostItem() == PostProc::coordinates )
   {
      GetCoordinateMatrix( coor_mat, ndim );
   }

   for ( int ip=1; ip<=get_nr_int_points(); ip++ )
   {
      total_ip_nr++;
      out << setw(7) << total_ip_nr;

      out.setf( ios::scientific, ios::floatfield );
      out.precision(3);

      if ( GetPostItem() == PostProc::coordinates )
      {
         Vector natcor( intlocs.get_coor(ip) );
         Matrix N( shape->get_N_vec( natcor ) );
         Vector X( ~(N*coor_mat) );
         out << " " << setw(11);
         X.print_row( out );
         out << endl;
      }
      else if ( GetPostItem() == PostProc::stresses )
      {
         out << " " << setw(11);
         (dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_xp_stress()).print_row(out);
         out << endl;
      }
      else if ( GetPostItem() == PostProc::eps_plast_eq )
      {
         out << " " << setw(11);
         (dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_xp_stress()).print_row(out);
         out << endl;
      }
   }
}

void NumIntStructElement::AddNodalPostContribution() const
{
   if ( GetPostItem() == PostProc::stresses )
   {
      Vector data( (dynamic_cast<Structural_Point*>(mat_point[0])->get_xp_stress()) );
      for ( int ip=2; ip<=get_nr_int_points(); ip++ )
      {
         data += (dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_xp_stress());
      }

      data *= 1.0/double(get_nr_int_points());

      for ( int i=1; i<=nr_of_element_nodes; i++ )
      {
         if ( Nodes[i-1]->dataContribution == 0 )
            Nodes[i-1]->dataVector = data;
         else
            Nodes[i-1]->dataVector += data;

         Nodes[i-1]->dataContribution++;
      }
   }
   else if ( GetPostItem() == PostProc::eps_plast_eq )
   {
      bool has_epq;
      double data( (dynamic_cast<Structural_Point*>(mat_point[0])->get_eps_plast_eq( has_epq )) );
      if ( has_epq )
      {
         for ( int ip=2; ip<=get_nr_int_points(); ip++ )
         {
            data += (dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_eps_plast_eq( has_epq ) );
         }

         data *= 1.0/double(get_nr_int_points());

         for ( int i=1; i<=nr_of_element_nodes; i++ )
         {
            if ( Nodes[i-1]->dataContribution == 0 )
               Nodes[i-1]->dataVector[0] = data;
            else
               Nodes[i-1]->dataVector[0] += data;

            Nodes[i-1]->dataContribution++;
         }
      }
   }
}

int NumIntStructElement::check_data( const Material_type *mat, const Geometry *geom,
                                     StressType st )
{
   int errors=0;

   switch ( st )
   {
      case psn:
      case axi:
      case gen: errors = check_mgd_ptrs( mat, geom, 1, 0 );
                break;
      case uni:
      case pss: errors = check_mgd_ptrs( mat, geom, 1, 1 );
                if ( geom != NULL && geom->get_Thick() <= 0.0 )
                {
                   cerr << "thickness <= 0; this is not allowed" << endl;
                   errors++;
                }
   }
   return errors;
}

NumIntThermalElement::NumIntThermalElement( int nr, Node_ptr *ElNodes, int nr_nodes,
                                            int nr_therm_dofs, int nr_dofs, Material_type *mat,
                                            Geometry *geom, StressType st,
                                            int nip ):NumIntElement()
{
   elm_nr = nr;
   Nodes = ElNodes;
   nr_of_element_nodes = nr_nodes;
   nr_of_local_dofs = nr_dofs;
   material_ptr = mat;
   geometry_ptr = geom;
   dimtype      = st;
   nr_thermal_dofs = nr_therm_dofs;

   mat_point = new MatPntPtr[nip];

   for ( int ip=0; ip<nip; ip++ )
      mat_point[ip] = mat->new_Material_Point( dimtype );
}

NumIntThermalElement::~NumIntThermalElement(){}

FiniteElement::LocalDof* NumIntThermalElement::GetLocalDofs() const
{
   LocalDof* ldofs;

   ldofs = new LocalDof[nr_thermal_dofs];

   int idof = 0;
   for ( int i=1; i<=nr_of_element_nodes; i++ )
   {
      ldofs[idof].phys_type = Dof::t;
      ldofs[idof].loc_node_nr = i;
      idof++;
   }

   if ( idof != nr_thermal_dofs )
      throw featured_error( "error in NumIntThermal::GetLocalDofs()" );

   return ldofs;
}

Matrix NumIntThermalElement::GetLocalStiff( Vector& /*stress_correction_vector*/ ) const
{
   int    ip, ndof;
   double dV, DetJ, delta_t, rho_c;
   Matrix D;

   ndof = nr_thermal_dofs;
   delta_t = GetTimeIncrement();

   Matrix Stifmat(ndof,ndof,0.0);

   IntLocs intlocs( get_int_locs() );

   int ndim( get_nodal_dimension( dimtype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      Matrix B( get_dNdX( intlocs.get_coor(ip), coor_mat, &DetJ ) );
      Matrix N( shape->get_N_vec( intlocs.get_coor(ip) ) );
      dV = DetJ*intlocs.get_weight( ip );

      D = dynamic_cast<Thermal_Point*>(mat_point[ip-1])->get_conduction_matrix();
      rho_c = dynamic_cast<Thermal_Point*>(mat_point[ip-1])->get_rho_c();

      Add_sAtBA( Stifmat, delta_t*dV, B, D );
      Stifmat += ~N*N*(rho_c*dV);
   }

   return Stifmat;
}

Vector NumIntThermalElement::GetLocalLHS() const
{
   int    ip, ndof;
   double dV, DetJ, dtime, rho_c_dT;
   Matrix B;
   Matrix N;
   Vector q;

   ndof = nr_thermal_dofs;

   Vector LHS(ndof,0);
   dtime = GetTimeIncrement();

   IntLocs intlocs( get_int_locs() );

   int ndim( get_nodal_dimension( dimtype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      B = get_dNdX( intlocs.get_coor(ip), coor_mat, &DetJ );
      N = shape->get_N_vec( intlocs.get_coor(ip) );
      dV = DetJ*intlocs.get_weight( ip );
      q = dynamic_cast<Thermal_Point*>(mat_point[ip-1])->get_flux();
      rho_c_dT = dynamic_cast<Thermal_Point*>(mat_point[ip-1])->get_accumulation();
      LHS += ~B*q*(dtime*dV) + ~N*(rho_c_dT*dV);
   }

   return LHS;
}

void NumIntThermalElement::iteration_update()
{
   IntLocs intlocs( get_int_locs() );

   Vector DT( get_local_increment_vector() );

   for ( int ip=1; ip<=get_nr_int_points(); ip++ )
      iteration_update_ip_therm( intlocs.get_coor(ip), DT,
                                 dynamic_cast<Thermal_Point*>(mat_point[ip-1]) );
}

void NumIntThermalElement::iteration_update_ip_therm( const Vector& natcor,
                                 const Vector& DT, Thermal_Point *matpoint )
{
      double DetJ;

   int ndim( get_nodal_dimension( dimtype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      Matrix B( get_dNdX( natcor, coor_mat, &DetJ ) );
      Vector dphi = B*DT;
      Matrix N( shape->get_N_vec( natcor ) );
      Matrix dT = N*DT;

      matpoint->set_flux_increment( dphi );
      matpoint->set_potential_increment( dT(1,1) );
}

Matrix NumIntThermalElement::get_dNdX( const Vector &natcor, const Matrix &coor,
                                       double* DetJ ) const
{
   double detj;

   Matrix dNdXi = shape->getPartialDerivatives( natcor );

   Matrix J = dNdXi*coor;
   detj = J.determinant();

   Matrix dNdX = J.inv()*dNdXi;

   if (      dimtype == psn )
      *DetJ = detj;
   else if ( dimtype == pss )
      *DetJ = detj*geometry_ptr->get_Thick();
   else if ( dimtype == axi )
   {
      Matrix N = shape->get_N_vec( natcor );
      Matrix x = N*coor;
      *DetJ = detj*PI*x(1,1);
   }
   else
      throw featured_error( "wrong stress type in NumIntThermalElement::get_dNdX()" );

   return dNdX;
}

void NumIntThermalElement::PrintStressHeader( ostream& out ) const
{
   int dim = get_nodal_dimension(dimtype);

   switch ( dim )
   {
      case 1: out << endl << "             q_x"                         << endl; break;
      case 2: out << endl << "             q_x         q_y"             << endl; break;
      case 3: out << endl << "             q_x         q_y         q_z" << endl;
   }
}

void NumIntThermalElement::PrintXpStressHeader( ostream& out ) const { PrintStressHeader( out ); }

void NumIntThermalElement::PrintStress( ostream& out ) const
{
   int    ip;

   out << "element " << elm_nr << endl;

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      out << setw(7) << ip;

      out.setf( ios::scientific, ios::floatfield );
      out.precision(3);
      out << " " << setw(11);
      (dynamic_cast<Thermal_Point*>(mat_point[ip-1])->get_flux()).print_row(out);
      out << endl;
   }
   out << endl;
}

void NumIntThermalElement::PrintXpStress( ostream& out ) const
{
   int    ip;

   out << "element " << elm_nr << endl;

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      out << setw(7) << ip;

      out.setf( ios::scientific, ios::floatfield );
      out.precision(3);
      out << " " << setw(11);
      (dynamic_cast<Thermal_Point*>(mat_point[ip-1])->get_flux()).print_row(out);
      out << endl;
   }
   out << endl;
}

NumIntThermoMechanicalElement::NumIntThermoMechanicalElement( int nr, Node_ptr *ElNodes,
                               int nr_nodes, int struct_dofs, int therm_dofs, Material_type *mat,
                               Geometry *geom, StressType st, int nip )
                              :NumIntStructElement( st ),NumIntThermalElement( st )
{
   elm_nr = nr;
   Nodes = ElNodes;
   nr_of_element_nodes = nr_nodes;
   nr_of_local_dofs = struct_dofs + therm_dofs;
   material_ptr = mat;
   geometry_ptr = geom;
   nr_struct_dofs = struct_dofs;
   nr_thermal_dofs = therm_dofs;

   mat_point = new MatPntPtr[nip];

   for ( int ip=0; ip<nip; ip++ )
      mat_point[ip] = mat->new_Material_Point( st );
}

NumIntThermoMechanicalElement::~NumIntThermoMechanicalElement(){}

FiniteElement::LocalDof* NumIntThermoMechanicalElement::GetLocalDofs() const
{
   LocalDof *ldofs, *struct_dofs, *thermal_dofs;

   ldofs = new LocalDof[nr_of_local_dofs];

   struct_dofs = NumIntStructElement::GetLocalDofs();
   thermal_dofs = NumIntThermalElement::GetLocalDofs();

   for ( int idof=0; idof<nr_struct_dofs; idof++ )
      ldofs[idof] = struct_dofs[idof];
   for ( int idof=0; idof<nr_thermal_dofs; idof++ )
      ldofs[idof+nr_struct_dofs] = thermal_dofs[idof];

   delete [] struct_dofs;
   delete [] thermal_dofs;

   if ( nr_struct_dofs + nr_thermal_dofs != nr_of_local_dofs )
      throw featured_error( "error in NumIntThermoMechanicalElement::GetLocalDofs()" );

   return ldofs;
}

Matrix NumIntThermoMechanicalElement::GetLocalStiff( Vector& stress_correction_vector ) const
{
   int ndof = nr_of_local_dofs;
   double DetJ, dV, alpha;
   Vector correction;
   bool expand = NumIntStructElement::NeedExpandStrain();

   Matrix Stifmat(ndof,ndof,0);

   if ( expand )
   {
      stress_correction_vector.resize(ndof);
      stress_correction_vector.clear();
   }

   Matrix StructMat( NumIntStructElement::GetLocalStiff( correction ) );
   stress_correction_vector.add_SubVec( 1, correction );

   Matrix ThermalMat( NumIntThermalElement::GetLocalStiff( correction ) );

   // coupling terms
   Matrix U_T_mat( nr_struct_dofs, nr_thermal_dofs, 0 );

   IntLocs intlocs( get_int_locs() );

   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   for ( int ip=1; ip<=get_nr_int_points(); ip++ )
   {
      Matrix B( get_B_mat( intlocs.get_coor(ip), coor_mat, &DetJ ) );
      Matrix N( shape->get_N_vec( intlocs.get_coor(ip) ) );

      dV = DetJ*intlocs.get_weight( ip );

      alpha = dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_expansion_coefficient();
      Vector D_alpha_dV( dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_D_times_iso_unit( alpha*dV ) );
      U_T_mat -= ~B*D_alpha_dV*N;
   }

   Stifmat.add_SubMat( 1, 1, StructMat );
   Stifmat.add_SubMat( nr_struct_dofs+1, nr_struct_dofs+1, ThermalMat );
   Stifmat.add_SubMat( 1, nr_struct_dofs+1, U_T_mat );

   return Stifmat;
}

Vector NumIntThermoMechanicalElement::GetLocalLHS() const
{
   int ndof = nr_of_local_dofs;
   Vector LHS( ndof, 0 );

   Vector StructLHS( NumIntStructElement::GetLocalLHS() );
   Vector ThermLHS( NumIntThermalElement::GetLocalLHS() );

   LHS.add_SubVec( 1, StructLHS );
   LHS.add_SubVec( nr_struct_dofs+1, ThermLHS );

   return LHS;
}

void NumIntThermoMechanicalElement::iteration_update()
{
   IntLocs intlocs( get_int_locs() );

   Vector dDof( get_local_increment_vector() );
   Vector dU( nr_struct_dofs );
   Vector DT( nr_thermal_dofs );
   for ( int i=1; i<=nr_struct_dofs;  i++ ) dU(i) = dDof(i);
   for ( int i=1; i<=nr_thermal_dofs; i++ ) DT(i) = dDof(i+nr_struct_dofs);

   for ( int ip=1; ip<=get_nr_int_points(); ip++ )
   {
      iteration_update_ip_therm( intlocs.get_coor(ip), DT,
                                 dynamic_cast<Thermal_Point*>(mat_point[ip-1]) );
      iteration_update_ip_struct( intlocs.get_coor(ip), dU,
                                  dynamic_cast<Structural_Point*>(mat_point[ip-1]) );
   }
}

void NumIntThermoMechanicalElement::PrintStress( ostream& out ) const
{
   NumIntStructElement::PrintStress( out );
}
void NumIntThermoMechanicalElement::PrintXpStress( ostream& out ) const
{
   NumIntStructElement::PrintXpStress( out );
}

NumInt2D::NumInt2D( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                    Material_type *mat, Geometry *geom, StressType st,
                    int nip )
                   :NumIntStructElement( nr, Nodes, nr_nodes, nr_dofs, nr_dofs, mat, geom,
                                         st, nip ){}

NumInt2D::~NumInt2D(){}

void NumInt2D::PrintStressHeader( ostream& out ) const
{
   if ( stype == axi )
      out << endl << "             Sxx         Syy         Szz         Sxy" << endl;
   else
      out << endl << "             Sxx         Syy         Sxy" << endl;
}

void NumInt2D::PrintXpStressHeader( ostream& out ) const
{
   out << endl << "             Sxx         Syy         Szz         Sxy" << endl;
}

void NumInt2D::PrintPostHeader( ostream &out, int stepnr ) const
{
   switch ( GetPostItem() )
   {
      case PostProc::coordinates:
                        out << "Gauss_points   0  " << setw(7) << get_nr_int_points()
                            << " 0   0   0" << endl;
                        break;
      case PostProc::stresses:
                        out << "stresses     " << ' ' << setw(7) << 1 << ' ' <<
                               setw(11) << stepnr << ' ' << setw(7) << 3 << ' ' <<
                               setw(7) << 1 << ' ' << setw(7)  << 1 << endl;
                        /* GiD wants Sxy at the third place for 2D results
                           keep consistent with mydomain.cpp */
                        out << "Sxx" << endl << "Syy" << endl << "Sxy" << endl
                            << "Szz" << endl;
                        break;
      case PostProc::eps_plast_eq:
                        out << "Eq_Pl_strain " << ' ' << setw(7) << 1 << ' ' <<
                               setw(11) << stepnr << ' ' << setw(7) << 1 << ' ' <<
                               setw(7) << 1 << ' ' << setw(7) << 1 << endl;
                        out << "Epeq" << endl;
                        break;
   }

   total_ip_nr = 0;
}

NumInt3D::NumInt3D( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                    Material_type *mat, Geometry *geom, int nip )
                   :NumIntStructElement( nr, Nodes, nr_nodes, nr_dofs, nr_dofs, mat, geom,
                                         gen, nip ){}

NumInt3D::~NumInt3D(){}

void NumInt3D::PrintStressHeader( ostream& out ) const
{
   out << endl << "             Sxx         Syy         Szz"
               <<     "         Sxy         Syz         Szx" << endl;
}

void NumInt3D::PrintXpStressHeader( ostream& out ) const
{
   PrintStressHeader( out );
}

void NumInt3D::PrintPostHeader( ostream &out, int stepnr ) const
{
   switch ( GetPostItem() )
   {
      case PostProc::coordinates:
                        out << "Gauss_points   0  " << setw(7) << get_nr_int_points()
                            << " 0   0   0" << endl;
                        break;
      case PostProc::stresses: out << "stresses     " << ' ' << setw(7) << 1 << ' ' <<
                               setw(11) << stepnr << ' ' << setw(7) << 3 << ' ' <<
                               setw(7) << 1 << ' ' << setw(7)  << 1 << endl;
                        out << "Sxx" << endl << "Syy" << endl << "Szz" << endl
                            << "Sxy" << endl << "Syz" << endl << "Szx" << endl;
                        break;
      case PostProc::eps_plast_eq:
                        out << "Eq_Pl_strain " << ' ' << setw(7) << 1 << ' ' <<
                               setw(11) << stepnr << ' ' << setw(7) << 1 << ' ' <<
                               setw(7) << 1 << ' ' << setw(7) << 1 << endl;
                        out << "Epeq" << endl;
                        break;
   }

   total_ip_nr = 0;
}

NumIntThermMech2D::NumIntThermMech2D( int nr, Node_ptr *Nodes, int nr_nodes, int struct_dofs,
                int therm_dofs,
                Material_type *mat, Geometry *geom, StressType st, int nip )
               :NumIntThermoMechanicalElement( nr, Nodes, nr_nodes, struct_dofs, therm_dofs,
                                               mat, geom, st, nip ){}

NumIntThermMech2D::~NumIntThermMech2D(){}

void NumIntThermMech2D::PrintStressHeader( ostream& out ) const
{
   if ( stype == axi )
      out << endl << "             Sxx         Syy         Szz         Sxy" << endl;
   else
      out << endl << "             Sxx         Syy         Sxy" << endl;
}

void NumIntThermMech2D::PrintXpStressHeader( ostream& out ) const
{
   out << endl << "             Sxx         Syy         Szz         Sxy" << endl;
}

Tri::Tri( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
          Material_type *mat, Geometry *geom, StressType st, int nip )
         :NumInt2D( nr, Nodes, nr_nodes, nr_dofs, mat, geom, st, nip ){}

Tri3::Tri3( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
            StressType st, int nip )
           :Tri( nr, Nodes, 3, 6, mat, geom, st, nip )
{
   shape = new FeaTure::Tri_lin(nip);
#ifdef FT_GUI
   draw_shape = new DrawTri_lin();
#endif
}

Tri6::Tri6( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
            StressType st, int nip, bool altnumber )
           :Tri( nr, Nodes, 6, 12, mat, geom, st, nip )
{
   if ( altnumber )
   {
      shape = new FeaTure::Tri_quad_alt(nip);
#ifdef FT_GUI
      draw_shape = new DrawTri_quad_alt();
#endif
   }
   else
   {
      shape = new FeaTure::Tri_quad(nip);
#ifdef FT_GUI
      draw_shape = new DrawTri_quad();
#endif
   }
}

Tri3therm::Tri3therm( int nr, Node_ptr *Nodes, Material_type *mat,
                      Geometry *geom )
                     :NumIntThermalElement( nr, Nodes, 3, 3, 3, mat, geom, psn, 1 )
{
   shape = new FeaTure::Tri_lin(1);
#ifdef FT_GUI
   draw_shape = new DrawTri_lin();
#endif
}

int Tri3therm::check_data( const Material_type *mat, const Geometry *geom )
{
   return check_mgd_ptrs( mat, geom, 1, 0 );
}

Quad::Quad( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
            Material_type *mat, Geometry *geom, StressType st,
            int nxi, int neta )
           :NumInt2D( nr, Nodes, nr_nodes, nr_dofs, mat, geom, st,
                      nxi*neta ){}

Quad4::Quad4( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
              StressType st, int nxi, int neta )
              :Quad( nr, Nodes, 4, 8, mat, geom, st, nxi, neta )
{
   shape = new FeaTure::Quad_lin(nxi,neta);
#ifdef FT_GUI
   draw_shape = new DrawQuad_lin();
#endif
}

Quad4ThermMech::Quad4ThermMech
            ( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
              StressType st, int nxi, int neta )
              :NumIntThermMech2D( nr, Nodes, 4, 8, 4, mat, geom, st, nxi*neta )
{
   shape = new FeaTure::Quad_lin(nxi,neta);
#ifdef FT_GUI
   draw_shape = new DrawQuad_lin();
#endif
}

Tri3ThermMech::Tri3ThermMech
            ( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
              StressType st, int nip )
              :NumIntThermMech2D( nr, Nodes, 3, 6, 3, mat, geom, st, nip )
{
   shape = new FeaTure::Tri_lin(nip);
#ifdef FT_GUI
   draw_shape = new DrawTri_lin();
#endif
}

Quad8::Quad8( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
              StressType st, int nxi, int neta, bool altnumber )
              :Quad( nr, Nodes, 8, 16, mat, geom, st, nxi, neta )
{
   if ( altnumber )
   {
      shape = new FeaTure::Quad_quad_alt(nxi,neta);
#ifdef FT_GUI
      draw_shape = new DrawQuad_quad_alt();
#endif
   }
   else
   {
      shape = new FeaTure::Quad_quad(nxi,neta);
#ifdef FT_GUI
      draw_shape = new DrawQuad_quad();
#endif
   }
}

Tri3psn::Tri3psn( int nr, Node_ptr *Nodes, Material_type *mat,
                  Geometry *geom )
                 :Tri3( nr, Nodes, mat, geom, psn, 1 ){}

Tri3pss::Tri3pss( int nr, Node_ptr *Nodes, Material_type *mat,
                  Geometry *geom )
                 :Tri3( nr, Nodes, mat, geom, pss, 1 ){}

Tri3axi::Tri3axi( int nr, Node_ptr *Nodes, Material_type *mat,
                  Geometry *geom )
                 :Tri3( nr, Nodes, mat, geom, axi, 1 ){}

Tri6psn::Tri6psn( int nr, Node_ptr *Nodes, Material_type *mat,
                  Geometry *geom, bool altnumber )
                 :Tri6( nr, Nodes, mat, geom, psn, 3, altnumber ){}

Tri6pss::Tri6pss( int nr, Node_ptr *Nodes, Material_type *mat,
                  Geometry *geom, bool altnumber )
                 :Tri6( nr, Nodes, mat, geom, pss, 3, altnumber ){}

Tri6axi::Tri6axi( int nr, Node_ptr *Nodes, Material_type *mat,
                  Geometry *geom, bool altnumber )
                 :Tri6( nr, Nodes, mat, geom, axi, 3, altnumber ){}

Quad4pss::Quad4pss( int nr, Node_ptr *Nodes, Material_type *mat,
                    Geometry *geom )
                   :Quad4( nr, Nodes, mat, geom, pss, 2, 2 ){}

Quad4psn::Quad4psn( int nr, Node_ptr *Nodes, Material_type *mat,
                    Geometry *geom )
                   :Quad4( nr, Nodes, mat, geom, psn, 2, 2 ){}

Quad4axi::Quad4axi( int nr, Node_ptr *Nodes, Material_type *mat,
                    Geometry *geom )
                   :Quad4( nr, Nodes, mat, geom, axi, 2, 2 ){}

Quad8pss::Quad8pss( int nr, Node_ptr *Nodes, Material_type *mat,
                    Geometry *geom, bool altnumber )
                   :Quad8( nr, Nodes, mat, geom, pss, 2, 2, altnumber ){}

Quad8psn::Quad8psn( int nr, Node_ptr *Nodes, Material_type *mat,
                    Geometry *geom, bool altnumber )
                   :Quad8( nr, Nodes, mat, geom, psn, 2, 2, altnumber ){}

Quad8axi::Quad8axi( int nr, Node_ptr *Nodes, Material_type *mat,
                    Geometry *geom, bool altnumber )
                   :Quad8( nr, Nodes, mat, geom, axi, 2, 2, altnumber ){}

Quad4pssTherm::Quad4pssTherm( int nr, Node_ptr *Nodes, Material_type *mat,
                              Geometry *geom )
                       :Quad4ThermMech( nr, Nodes, mat, geom, pss, 2, 2 ){}

Quad4psnTherm::Quad4psnTherm( int nr, Node_ptr *Nodes, Material_type *mat,
                              Geometry *geom )
                       :Quad4ThermMech( nr, Nodes, mat, geom, psn, 2, 2 ){}

Tri3psnTherm::Tri3psnTherm( int nr, Node_ptr *Nodes, Material_type *mat,
                            Geometry *geom )
                           :Tri3ThermMech( nr, Nodes, mat, geom, psn, 1 ){}

Hex::Hex( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
          Material_type *mat, Geometry *geom, int nxi, int neta, int nzeta )
         :NumInt3D( nr, Nodes, nr_nodes, nr_dofs, mat, geom, nxi*neta*nzeta ){}


Hex8::Hex8( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom )
           :Hex( nr, Nodes, 8, 24, mat, geom, 2, 2, 2 )
{
   shape = new FeaTure::Hex_lin(2,2,2);
#ifdef FT_GUI
   draw_shape = new DrawHex_lin();
#endif
}

int check_mgd_ptrs( const Material_type *mat, const Geometry *geom,
                    bool m, bool g )
{
   int error=0;

   if ( m && mat==NULL )
   {
      cerr << "no material group specified\n";
      cerr << "you must specify a material reference\n";
      error++;
   }

   if ( g && geom==NULL )
   {
      cerr << "no geometry group specified\n";
      cerr << "you must specify a geometry reference\n";
      error++;
   }

   return error;
}
