// Copyright (c) 1999, A.H. van den Boogaard

#include "tlelemen.h"
#include "constant.h"
#include "convert.h"
#include <cassert>
#include <iomanip>

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::CollectArray;
using FeaTure::Material_type;
using FeaTure::Structural_Point;
using FeaTure::Geometry;
using FeaTure::StressType;
using FeaTure::psn;
using FeaTure::pss;
using FeaTure::axi;
using FeaTure::gen;
using FeaTure::IntLocs;
using FeaTure::get_nodal_dimension;

using std::endl;
using std::ostream;

TLNumIntStructElement::TLNumIntStructElement( int nr, Node_ptr *Nodes,
                       int nr_nodes, int struct_dofs,
                       int nr_dofs, Material_type *mat, Geometry *geom,
                       StressType st, int nip )
                       :NumIntStructElement( nr, Nodes, nr_nodes, struct_dofs,
                        nr_dofs, mat, geom, st, nip )
{
   defgrad0 = new Tensor2Gen [nip];
   defgrad1 = new Tensor2Gen [nip];
   for ( int i=0; i<nip; i++ )
   {
      defgrad0[i] = 1;
      defgrad1[i] = 1;
   }
}

TLNumIntStructElement::~TLNumIntStructElement()
{
   delete [] defgrad0;
   delete [] defgrad1;
}

void TLNumIntStructElement::WriteElementRestartData( ostream& restart ) const
{
   NumIntElement::WriteElementRestartData( restart );
   CollectArray data;

   for ( int ip=0; ip<get_nr_int_points(); ip++ )
      Tensor2CollectArray( defgrad0[ip], &data );

   restart << data;
}

void TLNumIntStructElement::ReadElementRestartData( std::istream& restart )
{
   NumIntElement::ReadElementRestartData( restart );
   CollectArray data;
   restart >> data;

   for ( int ip=0; ip<get_nr_int_points(); ip++ )
   {
      CollectArray2Tensor( data, &defgrad0[ip] );
      defgrad1[ip] = defgrad0[ip];
   }
}

void TLNumIntStructElement::make_BL_mat_ps( const Matrix &dNdX,
                                            const Tensor2Gen &F1,
                                            Matrix &BL ) const
{
/* Determine large deformation B-matrix for plane stress and plane strain */

   double F11, F12, F21, F22;
   F11 = F1(1,1); F12 = F1(1,2);
   F21 = F1(2,1); F22 = F1(2,2);

   for ( int i=1; i<=nr_of_element_nodes; i++ )
   {
      BL(1,2*i-1) = F11*dNdX(1,i);
      BL(1,2*i)   = F21*dNdX(1,i);
      BL(2,2*i-1) = F12*dNdX(2,i);
      BL(2,2*i)   = F22*dNdX(2,i);
      BL(3,2*i-1) = F11*dNdX(2,i)+F12*dNdX(1,i);
      BL(3,2*i)   = F21*dNdX(2,i)+F22*dNdX(1,i);
   }
}

void TLNumIntStructElement::make_BL_mat_gen( const Matrix &dNdX,
                                             const Tensor2Gen &F1,
                                             Matrix &BL ) const
{
/* Determine large deformation B-matrix for 3D situations */

   double F11, F12, F13, F21, F22, F23, F31, F32, F33;
   F11 = F1(1,1); F12 = F1(1,2), F13 = F1(1,3);
   F21 = F1(2,1); F22 = F1(2,2); F23 = F1(2,3);
   F31 = F1(3,1); F32 = F1(3,2); F33 = F1(3,3);

   for ( int i=1; i<=nr_of_element_nodes; i++ )
   {
      BL(1,3*i-2) = F11*dNdX(1,i);
      BL(1,3*i-1) = F21*dNdX(1,i);
      BL(1,3*i  ) = F31*dNdX(1,i);
      BL(2,3*i-2) = F12*dNdX(2,i);
      BL(2,3*i-1) = F22*dNdX(2,i);
      BL(2,3*i  ) = F32*dNdX(2,i);
      BL(3,3*i-2) = F13*dNdX(3,i);
      BL(3,3*i-1) = F23*dNdX(3,i);
      BL(3,3*i  ) = F33*dNdX(3,i);
      BL(4,3*i-2) = F11*dNdX(2,i)+F12*dNdX(1,i);
      BL(4,3*i-1) = F21*dNdX(2,i)+F22*dNdX(1,i);
      BL(4,3*i  ) = F31*dNdX(2,i)+F32*dNdX(1,i);
      BL(5,3*i-2) = F12*dNdX(3,i)+F13*dNdX(2,i);
      BL(5,3*i-1) = F22*dNdX(3,i)+F23*dNdX(2,i);
      BL(5,3*i  ) = F32*dNdX(3,i)+F33*dNdX(2,i);
      BL(6,3*i-2) = F11*dNdX(3,i)+F13*dNdX(1,i);
      BL(6,3*i-1) = F21*dNdX(3,i)+F23*dNdX(1,i);
      BL(6,3*i  ) = F31*dNdX(3,i)+F33*dNdX(1,i);
   }
}

void TLNumIntStructElement::make_BL_mat_axi( const Matrix &dNdX,
                                             const Tensor2Gen &F1,
                                             double r, const Matrix &N,
                                             Matrix &BL ) const
{
/* Determine large deformation B-matrix for 2D axi-symmetric situations */

   double F11, F12, F21, F22, F33, rinv;
   F11 = F1(1,1); F12 = F1(1,2);
   F21 = F1(2,1); F22 = F1(2,2); F33 = F1(3,3);

   assert( r >= 1.e-20 );
   /* radius is always > 0, therefore abs() is not necessary */
   rinv = 1.0/r;

   for ( int i=1; i<=nr_of_element_nodes; i++ )
   {
      BL(1,2*i-1) = F11*dNdX(1,i);
      BL(1,2*i)   = F21*dNdX(1,i);
      BL(2,2*i-1) = F12*dNdX(2,i);
      BL(2,2*i)   = F22*dNdX(2,i);
      BL(3,2*i-1) = F33*rinv*N(1,i);
      BL(4,2*i-1) = F11*dNdX(2,i)+F12*dNdX(1,i);
      BL(4,2*i)   = F21*dNdX(2,i)+F22*dNdX(1,i);
   }
}

Matrix TLNumIntStructElement::get_BL_mat( const Vector &natcor,
                                          const Matrix &coor_mat,
                                          const Tensor2Gen &F1,
                                          double* DetJ ) const
{
   int    nstr, ndim, ndof;

   nstr = StressSize( stype );
   ndim = get_nodal_dimension( stype );
   ndof = ndim*nr_of_element_nodes;

   Matrix dNdX = get_dNdX( natcor, coor_mat, DetJ );

   Matrix BL(nstr,ndof,0.0);

   if ( stype == psn || stype == pss )
   {
      if ( stype==pss ) *DetJ *= geometry_ptr->get_Thick();

      make_BL_mat_ps( dNdX, F1, BL );
   }
   else if ( stype == axi )
   {
      Matrix N = shape->get_N_vec( natcor );
      Matrix x = N*coor_mat;
      *DetJ *= PI2*x(1,1);

      make_BL_mat_axi( dNdX, F1, x(1,1), N, BL );
   }
   else if ( stype == gen )
   {
      make_BL_mat_gen( dNdX, F1, BL );
   }
   else
   {
      std::cerr << "wrong stress type in TLNumIntStructElement::get_BL_mat()"
                << endl;
      throw( EXIT_FAILURE );
   }

   return BL;
}

void TLNumIntStructElement::get_TL_B_mats( const Vector &natcor,
                                           const Matrix &coor_mat,
                                           const Tensor2Gen &F1,
                                           Matrix &BL, Matrix &BNL,
                                           double* DetJ ) const
{
   int    nstr, ndof, ndim;

   nstr = StressSize( stype );
   ndim = get_nodal_dimension( stype );
   ndof = ndim*nr_of_element_nodes;

   Matrix dNdX = get_dNdX( natcor, coor_mat, DetJ );

   BL.resize(nstr,ndof);
   BL.clear();


   if ( stype == psn || stype == pss )
   {
      if ( stype==pss ) *DetJ *= geometry_ptr->get_Thick();

      make_BL_mat_ps( dNdX, F1, BL );

      BNL.resize(ndim*ndim,ndof);
      BNL.clear();
      for ( int i=1; i<=nr_of_element_nodes; i++ )
      {
         BNL(1,2*i-1) = dNdX(1,i);
         BNL(2,2*i-1) = dNdX(2,i);
         BNL(3,2*i)   = dNdX(1,i);
         BNL(4,2*i)   = dNdX(2,i);
      }
   }
   else if ( stype == axi )
   {
      Matrix N = shape->get_N_vec( natcor );
      Matrix x = N*coor_mat;
      *DetJ *= PI2*x(1,1);

      make_BL_mat_axi( dNdX, F1, x(1,1), N, BL );

      /* make_BL_mat_axi also asserts that x(1,1) != 0.0 */
      BNL.resize(5,ndof);
      BNL.clear();
      for ( int i=1; i<=nr_of_element_nodes; i++ )
      {
         BNL(1,2*i-1) = dNdX(1,i);
         BNL(2,2*i-1) = dNdX(2,i);
         BNL(3,2*i)   = dNdX(1,i);
         BNL(4,2*i)   = dNdX(2,i);
         BNL(5,2*i-1) = N(1,i)/x(1,1);
      }
   }
   else if ( stype == gen )
   {
      make_BL_mat_ps( dNdX, F1, BL );

      BNL.resize(ndim*ndim,ndof);
      BNL.clear();
      for ( int i=1; i<=nr_of_element_nodes; i++ )
      {
         BNL(1,2*i-2) = dNdX(1,i);
         BNL(2,2*i-2) = dNdX(2,i);
         BNL(3,2*i-2) = dNdX(3,i);
         BNL(4,2*i-1) = dNdX(1,i);
         BNL(5,2*i-1) = dNdX(2,i);
         BNL(6,2*i-1) = dNdX(3,i);
         BNL(7,2*i)   = dNdX(1,i);
         BNL(8,2*i)   = dNdX(2,i);
         BNL(9,2*i)   = dNdX(3,i);
      }
   }
   else
   {
      std::cerr << "wrong stress type in TLNumInt2D::get_TL_B_mats()" << endl;
      throw( EXIT_FAILURE );
   }
}

Matrix TLNumIntStructElement::GetLocalStiff(
                                      Vector& stress_correction_vector ) const
{
   int    ip, ndof, ndim, ndim2;
   double dV, DetJ;
   Matrix D;

   bool expand = NeedExpandStrain();

   ndof = nr_struct_dofs;
   ndim = get_nodal_dimension( stype );
   assert( ndim>=1 && ndim <=3 );
   ndim2 = ndim*ndim;

   Matrix Stifmat(ndof,ndof,0);

   if ( expand )
   {
      stress_correction_vector.resize(ndof);
      stress_correction_vector.clear();
   }

   IntLocs intlocs( get_int_locs() );

   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      Matrix BL, BNL;
      get_TL_B_mats( intlocs.get_coor(ip), coor_mat, defgrad1[ip-1],
                     BL, BNL, &DetJ );
      dV = DetJ*intlocs.get_weight( ip );

      Matrix T( ndim2, ndim2, 0.0 );
      Vector S( dynamic_cast<Structural_Point*>(mat_point[ip-1])->
                                                                get_stress() );

      if ( ndim == 1 )
      {
         T.resize(1,1);
         T(1,1) = S(1);
      }
      else if ( ndim == 2 )
      {
         if ( stype == pss || stype == psn )
            T.resize(4,4);
         else
            T.resize(5,5);
         T.clear();
         T(1,1) = S(1);
         T(1,2) = T(2,1) = stype == axi ? S(4) : S(3);
         T(2,2) = S(2);

         for ( int i=0; i<2; i++ )
            for ( int j=0; j<2; j++ )
               T[i+2][j+2] = T[i][j];

         if ( stype == axi ) T(5,5) = S(3);
      }
      else if ( ndim == 3 )
      {
         T.resize(9,9);
         T.clear();
         T(1,1) = S(1); T(1,2) = S(4); T(1,3) = S(6);
         T(2,1) = S(4); T(2,2) = S(2); T(2,3) = S(5);
         T(3,1) = S(6); T(3,2) = S(5); T(3,3) = S(3);
         for ( int i=0; i<3; i++ )
            for ( int j=0; j<3; j++ )
            {
               T[i+3][j+3] = T[i][j];
               T[i+6][j+6] = T[i][j];
            }
      }

      if ( expand )
      {
         Vector stress_correction;
         dynamic_cast<FeaTure::Structural_Point*>(mat_point[ip-1])->
                           get_D_and_stress_correction( D, stress_correction );
         stress_correction_vector -= ~BL*stress_correction *dV;
      }
      else D = dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_D_mat();

//      Stifmat += (~BL*D*BL + ~BNL*T*BNL )*dV;
      Add_sAtBA( Stifmat, dV, BL, D );
      Add_sAtBA( Stifmat, dV, BNL, T );
   }

   return Stifmat;
}

Vector TLNumIntStructElement::GetLocalLHS() const
{
   int    ip, ndof;
   double dV, DetJ;
   Matrix B;
   Vector sigma;

   ndof = nr_struct_dofs;

   Vector LHS(ndof,0);

   IntLocs intlocs( get_int_locs() );

   int ndim( get_nodal_dimension( stype ) );
   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      B = get_BL_mat( intlocs.get_coor(ip), coor_mat, defgrad1[ip-1], &DetJ );
      dV = DetJ*intlocs.get_weight( ip );
      sigma = dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_stress();
      LHS += ~B*sigma *dV;
   }

   return LHS;
}

void TLNumIntStructElement::iteration_update()
{
   int ndim( get_nodal_dimension( stype ) );

   IntLocs intlocs( get_int_locs() );

   Vector dU( get_local_increment_vector() );

   Vector du, dv, dw;
   du.resize( nr_of_element_nodes );
   if ( ndim >= 2 ) dv.resize( nr_of_element_nodes );
   if ( ndim == 3 ) dw.resize( nr_of_element_nodes );
   for ( int i=0; i<nr_of_element_nodes; i++ )
   {
      du[i] = dU[ndim*i];
      if ( ndim >= 2 ) dv[i] = dU[ndim*i+1];
      if ( ndim == 3 ) dw[i] = dU[ndim*i+2];
   }

   Matrix coor_mat;
   GetCoordinateMatrix( coor_mat, ndim );

   for ( int ip=1; ip<=get_nr_int_points(); ip++ )
   {
      double DetJ;

      Matrix dNdX( get_dNdX( intlocs.get_coor(ip), coor_mat, &DetJ ) );

      Vector dF1, dF2, dF3;
      dF1 = dNdX*du;
      if ( ndim >= 2 ) dF2 = dNdX*dv;
      if ( ndim == 3 ) dF3 = dNdX*dw;

      Tensor2Gen DF;
      DF = 0;
      DF(1,1) = dF1(1);
      if ( ndim >= 2 )
      {
         DF(1,2) = dF1(2);
         DF(2,1) = dF2(1);
         DF(2,2) = dF2(2);
         if ( ndim == 3 )
         {
            DF(1,3) = dF1(3);
            DF(2,3) = dF2(3);
            DF(3,1) = dF3(1);
            DF(3,2) = dF3(2);
            DF(3,3) = dF3(3);
         }
      }

      defgrad1[ip-1] = defgrad0[ip-1] + DF;
      Tensor2Sym Dstrain( sym(defgrad0[ip-1].transpose()*DF)+0.5*multAtA(DF) );

      Vector dstrain( Tensor2StrainVector( Dstrain, false, stype ) );

      dynamic_cast<Structural_Point*>(mat_point[ip-1])->
                                               set_strain_increment( dstrain );
   }
}

void TLNumIntStructElement::finish_increment()
{
   for ( int ip=0; ip<get_nr_int_points(); ip++ )
   {
      defgrad0[ip] = defgrad1[ip];
      mat_point[ip]->finish_increment();
   }
}

Matrix TLNumIntStructElement::get_dNdX( const Vector &natcor,
                                        const Matrix &coor,
                                        double* DetJ ) const
{
   Matrix dNdXi = shape->getPartialDerivatives( natcor );

   Matrix J = dNdXi*coor;
   *DetJ = J.determinant();

   return J.inv()*dNdXi; // = dNdX
}

void TLNumIntStructElement::PrintXpStress( ostream& out ) const
{
   int    ip;

   out << "element " << elm_nr << endl;

   for ( ip=1; ip<=get_nr_int_points(); ip++ )
   {
      out << std::setw(7) << ip;

      out.setf( std::ios::scientific, std::ios::floatfield );
      out.precision(3);
      out << " " << std::setw(11);
      Vector sigma(
            dynamic_cast<Structural_Point*>(mat_point[ip-1])->get_xp_stress());
      Tensor2Sym S( StressVector2Tensor( sigma, true, stype ) );

      Tensor2Sym Cauchy( det(defgrad1[ip-1]) * multASAt(defgrad1[ip-1],S) );

      sigma = Tensor2StressVector( Cauchy, true, stype );
      sigma.print_row(out);
      out << endl;
   }
   out << endl;
}

TLNumInt2D::TLNumInt2D( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                      Material_type *mat, Geometry *geom, StressType st,
                      int nip )
                     :TLNumIntStructElement( nr, Nodes, nr_nodes, nr_dofs,
                                             nr_dofs, mat, geom, st, nip ){}

TLNumInt2D::~TLNumInt2D(){}

void TLNumInt2D::PrintStressHeader( ostream& out ) const
{
   if ( stype == axi )
      out << endl
          << "             Sxx         Syy         Szz         Sxy"
          << endl;
   else
      out << endl
          << "             Sxx         Syy         Sxy" << endl;
}

void TLNumInt2D::PrintXpStressHeader( ostream& out ) const
{
   out << endl
       << "             Sxx         Syy         Szz         Sxy"
       << endl;
}

TLQuad::TLQuad( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                Material_type *mat, Geometry *geom, StressType st,
                int nxi, int neta )
               :TLNumInt2D( nr, Nodes, nr_nodes, nr_dofs, mat, geom, st,
                            nxi*neta ){}

TLQuad4::TLQuad4( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
                  StressType st, int nxi, int neta )
                  :TLQuad( nr, Nodes, 4, 8, mat, geom, st, nxi, neta )
{
   shape = new FeaTure::Quad_lin(nxi,neta);
#ifdef FT_GUI
   draw_shape = new DrawQuad_lin();
#endif
}

TLQuad8::TLQuad8( int nr, Node_ptr *Nodes, Material_type *mat, Geometry *geom,
                  StressType st, int nxi, int neta )
                  :TLQuad( nr, Nodes, 8, 16, mat, geom, st, nxi, neta )
{
   shape = new FeaTure::Quad_quad(nxi,neta);
#ifdef FT_GUI
   draw_shape = new DrawQuad_quad();
#endif
}

TLQuad4psn::TLQuad4psn( int nr, Node_ptr *Nodes, Material_type *mat,
                        Geometry *geom )
                       :TLQuad4( nr, Nodes, mat, geom, psn, 2, 2 ){}

TLQuad8psn::TLQuad8psn( int nr, Node_ptr *Nodes, Material_type *mat,
                        Geometry *geom )
                       :TLQuad8( nr, Nodes, mat, geom, psn, 2, 2 ){}

