// Copyright (c) 1999, A.H. van den Boogaard

#include "control.h"

#include <fstream>
#include <cstdio>
#include <iomanip>

template <class DomainT>
void FeaTure::Control<DomainT>::solve( DomainT& model, std::ostream &out,
                                       std::ostream &post )
{
   initialize( model, out );
   continuate( model, out, post, -1 );
}

template <class DomainT>
FeaTure::IncrementalControl<DomainT>::IncrementalControl():Control<DomainT>()
{
   incr_nr=0;
   max_incr = 1;

   time_increment = 0.0;
   time0 = time1  = 0.0;

   use_damping = false;

   Rayleigh_alpha = 0.0;
   Rayleigh_beta  = 0.0;

   startup = start;

   max_resize = 0;
   this->file_basename = "restart";
}

template <class DomainT>
void FeaTure::IncrementalControl<DomainT>::set_increments( int set,
                                                        double size, int nr )
{
   LoadBlock load_block;
   load_block.load_set = set;
   load_block.size = size;
   load_block.nr_increments = nr;
   
   load_blocks.push_back( load_block );
}

template <class DomainT>
void FeaTure::IncrementalControl<DomainT>::set_time_increment( double dtime )
{
   time_increment = dtime;
}

template <class DomainT>
float FeaTure::IncrementalControl<DomainT>::get_fraction_executed( int flag )
{
   if ( flag == STEPS ) return ( float(incr_nr) / float(max_incr) );
   else return 0.0;
}

template <class DomainT>
void FeaTure::IncrementalControl<DomainT>::set_Rayleigh_damping( double alpha,
                                                                 double beta )
{
   use_damping = true;
   Rayleigh_alpha = alpha;
   Rayleigh_beta  = beta;
}

template <class DomainT>
void FeaTure::IncrementalControl<DomainT>::init_incremental( DomainT& model )
{
   int    nr_free;

   nr_free = model.get_nr_free();

   incr_nr = 0;

   rhs.resize( nr_free );

   if ( startup == restart ) ReadRestartData( model );

   max_incr = incr_nr;  // incr_nr may have been set by Restart

   for ( typename std::deque<LoadBlock>::iterator lb=load_blocks.begin();
                                                  lb!=load_blocks.end(); lb++ )
      max_incr += (*lb).nr_increments;
}

template <class DomainT>
bool FeaTure::IncrementalControl<DomainT>::continuate( DomainT& model,
                                         std::ostream &out, std::ostream &post,
                                              int steps )
{
/*
   steps:  nr of steps to take before returning, if negative finish until end
           used to return control to caller e.g. to GUI in same thread
*/
   bool succes;
   int  loadset, current_steps=0;

   while ( (!load_blocks.empty()) && ( current_steps<steps || steps<0 ) )
   {
      loadset = load_blocks[0].load_set;
      double inisize = load_blocks[0].size;
   
      ElementBase::SetTimeIncrement( time_increment );
      Material_Point::SetTimeIncrement( time_increment );
   
      while ( ( steps<0 || current_steps<steps ) &&
                load_blocks[0].nr_increments > 0    )
      {
         double size = inisize;
         double factor=1.0;
         int    restartnr=0;
         incr_nr++;
         current_steps++;
         (load_blocks[0].nr_increments)--;
   
         time0 = time1;
         log_increment_nr( out );
 
         // resize loop
         do
         {
            try
            {
               time1 = time0 + time_increment;
               solve_increment( model, out, loadset, size );

               if ( this->postproc.process( incr_nr, is_last() ) )
                  model.post_file( post, incr_nr, this->postproc );

               if ( is_last() ) model.print_results( out );
   
               model.finish_increment();

               if ( save_restart_file.is_selected( incr_nr, is_last() ) )
                  WriteRestartData( model );
   
               succes = true;
   
               this->has_run = true;
            }
            catch ( const not_converged_error& div )
            {
               succes = false;
               if ( restartnr < max_resize )
               {
                  size /= 2.0;
                  time_increment /= 2.0;
                  factor /= 2.0;
                  restartnr++;
 
                  out  << std::endl << "Resized because " << div.what() <<
                                  ", factor " << factor << std::endl;
                  std::cout << std::endl << "Resized because " << div.what() <<
                                  ", factor " << factor << std::endl;
               }
               else
               {
                  out  << std::endl << "ERROR: " << div.what() << std::endl;
                  std::cout << std::endl << "ERROR: " << div.what() << std::endl;
                  throw feature_error( "aborted increment" );
               }
            }
         } while ( !succes );
      }
   
      if ( load_blocks[0].nr_increments == 0 ) load_blocks.pop_front();

   }
   return !load_blocks.empty();
}

template <class DomainT>
void FeaTure::IncrementalControl<DomainT>::WriteRestartData( DomainT& model )
{
   char *filename;
   filename = new char[this->file_basename.size()+15];

   if ( incr_nr < 1000 )
      // zero-padded: 001, 002 etc.
      std::sprintf( filename, "%s%03u.bin", this->file_basename.c_str(), incr_nr );
   else
      std::sprintf( filename, "%s%u.bin", this->file_basename.c_str(), incr_nr );

#ifndef __GNUG__
   std::ofstream restartfile( filename, std::ios::binary );
#else
   std::ofstream restartfile( filename );
#endif

   delete filename;

   restartfile << std::setprecision(15);
   restartfile << "feature0.1restart" << std::endl;
   restartfile << incr_nr << std::endl;
   restartfile << time1 << std::endl;
   model.WriteRestartData( restartfile );
   restartfile << "end_restart_file" << std::endl;
}

template <class DomainT>
void FeaTure::IncrementalControl<DomainT>::ReadRestartData( DomainT& model )
{
   std::string marker;

#ifndef __GNUG__
   std::ifstream restartfile( restartfilename.c_str(), std::ios::binary );
#else
   std::ifstream restartfile( restartfilename.c_str() );
#endif

#ifdef __GNUG__
   char tmp[50];
   restartfile.get( tmp, 50, '\n' );
   marker = tmp;
#else
   restartfile >> marker;
#endif
   if ( marker != "feature0.1restart" )
      throw file_error( "incorrect restart file" );

   restartfile >> incr_nr;
   restartfile >> time1;
   model.ReadRestartData( restartfile );

   restartfile >> marker;
   if ( marker != "end_restart_file" )
      throw file_error( "incorrect end of restart file" );

   this->has_run = true;
}

template <class DomainT>
void FeaTure::IncrementalControl<DomainT>::log_increment_nr( std::ostream &out )
{
   std::cout << std::endl << "increment " << std::setw(3) << incr_nr << std::endl;
   if ( std::cout.rdbuf() != out.rdbuf() )
   {
      out << std::endl << "increment " << std::setw(3) << incr_nr << std::endl;
   }
}

template <class DomainT>
FeaTure::IncrementalIterativeControl<DomainT>::IncrementalIterativeControl()
   :IncrementalControl<DomainT>()
{
   max_iter    = 10;
   EPS_UNB     = 1.E-4;
   load_norm   = 0;
   unbalance_norm = 0;
}

template <class DomainT>
FeaTure::IncrementalIterativeControl<DomainT>::~IncrementalIterativeControl()
{}

template <class DomainT>
void FeaTure::IncrementalIterativeControl<DomainT>::set_iteration_method(
                                                                   int method )
{
   iteration_method = method;
}

template <class DomainT>
void FeaTure::IncrementalIterativeControl<DomainT>::set_max_iter( int maxiter )
{
   max_iter = maxiter;
}

template <class DomainT>
void FeaTure::IncrementalIterativeControl<DomainT>::set_unbalance_crit(
                                                                  double eps )
{
   EPS_UNB = eps;
}

template <class DomainT>
void FeaTure::IncrementalIterativeControl<DomainT>::calculate_norms()
{
   unbalance_norm = this->rhs.norm();
}

template <class DomainT>
void FeaTure::IncrementalIterativeControl<DomainT>::solve_iteration(
                               SystemMat &Systemptr, DomainT& model )
{
   Vector solution;
   Systemptr.solve( this->rhs, solution );
   model.add_to_increments( solution );
   model.iteration_update();
}

template <class DomainT>
bool FeaTure::IncrementalIterativeControl<DomainT>::converged() const
{
   if ( load_norm > 1E-12 )
   {
      double ratio( unbalance_norm / load_norm );

      if ( ratio <= EPS_UNB ) return true;
      else if ( ratio > 1.0e10 ) throw divergence_error( "divergence occured" );
      else return false;
   }
   else return unbalance_norm < 1E-10;
}
