// Copyright (c) 2003, A.H. van den Boogaard

#ifndef FTMATRIX_H
#define FTMATRIX_H

#include <cmath>
#include <cassert>
#include <iostream>

namespace FeaTure
{

class Matrix;  // forward declaration

class Vector
{
   private:
      size_t m_;
      double *v_;

   public:
      Vector() : m_(0),v_(0){}

      Vector(size_t N) : m_(N)
      {
         // does not initialize data!
         v_ = new double[m_];
      }

      Vector(size_t N, double V) : m_(N)
      {
         v_ = new double[m_];
         for (size_t i=0; i<m_; i++) v_[i]=V;
      }

      Vector(const Vector &a) : m_(a.m_)
      {
         v_ = new double[m_];
         for (size_t i=0; i<m_; i++) v_[i] = a.v_[i];
      }
 
      // not every matrix is a vector, therefore an explicit constructor

      explicit Vector(const Matrix &A);

      ~Vector(){ delete [] v_; }

      Vector& operator=(const Vector &a)
      {
         resize(a.m_);
         for (size_t i=0; i<m_; i++) v_[i] = a.v_[i];
         return *this;
      }
 
      size_t size() const { return m_; }

      double& operator[](size_t i)
      {
         //assert(0<=i);  // useless for a size_t, but not for int
         assert(i<m_);
         return v_[i];
      }
 
      const double& operator[](size_t i) const
      {
         //assert(0<=i);  // useless for a size_t, but not for int
         assert(i<m_);
         return v_[i];
      }

      double& operator()(size_t i)
      {
         assert(1<=i);
         assert(i<= m_);
         return v_[i-1];
      }
 
      const double& operator() (size_t i) const
      {
         assert(1<=i);
         assert(i<= m_);
         return v_[i-1];
      }

      void clear(){ for ( size_t i=0; i<m_; i++ ) v_[i]=0.0; }
      void resize( size_t row );

      Vector add_SubVec( size_t row, const Vector& SubVec );

      double norm() const
      {
         double sqnorm=0;
         for ( size_t i=0; i<m_; i++ ) sqnorm += v_[i]*v_[i];
         return std::sqrt(sqnorm);
      }

      void print_row(std::ostream& os) const;

      Vector& operator*=(const double &B);
      Vector& operator+=(const Vector &B);
      Vector& operator-=(const Vector &B);

      //double* fortran_vector(){ return v_; }

      friend Vector mult( const Matrix &A, const Vector &b );
      friend Vector tmult( const Matrix &A, const Vector &b );
      friend double inner_product(const Vector &a, const Vector &b, double v);
};

class Matrix
{
   private:
      size_t m_, n_;
      double *v_;

   public:
      Matrix() : m_(0), n_(0), v_(0){}

      Matrix(size_t M, size_t N ) : m_(M), n_(N)
      {
         // does not initialize data!
         v_ = new double[m_*n_];
      }
 
      Matrix(size_t M, size_t N, double V) : m_(M), n_(N)
      {
         size_t mn(m_*n_);
         v_ = new double[mn];
         for (size_t i=0; i<mn; i++) v_[i] = V;
      }
 
      Matrix(const Matrix &A) : m_(A.m_), n_(A.n_)
      {
         size_t mn(m_*n_);
         v_ = new double[mn];
         for (size_t i=0; i<mn; i++) v_[i] = A.v_[i];
      }
 
      Matrix(const Vector &a) : m_(a.size()), n_(1)
      {
         v_ = new double[m_];
         for (size_t i=0; i<m_; i++) v_[i] = a[i];
      }
 
      ~Matrix()
      {
         delete [] v_;
      }

      Matrix& operator=(const Matrix &A)
      {
         resize(A.m_,A.n_);
         size_t mn(m_*n_);
         for (size_t i=0; i<mn; i++) v_[i] = A.v_[i];
         return *this;
      }
 
      size_t size1() const { return m_; }
      size_t size2() const { return n_; }

      double* operator[](size_t i)
      {
         //assert(0<=i);  // useless for a size_t, but not for int
         assert(i<m_);
         return &v_[i*n_];
      }
 
      const double* operator[](size_t i) const
      {
         //assert(0<=i);  // useless for a size_t, but not for int
         assert(i< m_);
         return &v_[i*n_];
      }

      double& operator()(size_t i, size_t j)
      {
         assert(1<=i);
         assert(i<= m_);
         assert(1<=j);
         assert(j<= n_);
         return v_[(i-1)*n_+j-1];
      }
 
      const double& operator() (size_t i, size_t j) const
      {
         assert(1<=i);
         assert(i<= m_);
         assert(1<=j);
         assert(j<= n_);
         return v_[(i-1)*n_+j-1];
      }

      void clear(){ for ( size_t i=0; i<m_*n_; i++ ) v_[i]=0.0; }
      void resize(size_t row, size_t col);

      Matrix without_rowcol( size_t rowcol ) const;
      Matrix add_SubMat( size_t row, size_t col, const Matrix& SubMat );

      double determinant() const;
      Matrix inv() const;

      Matrix& operator*=(const double &B);
      Matrix& operator+=(const Matrix &B);
      Matrix& operator-=(const Matrix &B);

      //double* fortran_Tmatrix(){ return v_; }

      friend Matrix mult(const Matrix &A, const Matrix &B);
      friend Vector mult(const Matrix &A, const Vector &b);
      friend Matrix tmult(const Matrix &A, const Matrix &B);
      friend Matrix tmult(const Vector &a, const Matrix &B);
      friend Vector tmult(const Matrix &A, const Vector &b);
      friend void Add_sAtBA(Matrix &C, double s, const Matrix &A,
                            const Matrix &B);
};

// definition of member functions

inline Vector::Vector(const Matrix &A) : m_(A.size1())
{
   assert( A.size2()==1 );
   v_ = new double[m_];
   for (size_t i=0; i<m_; i++) v_[i] = A[i][0];
}

inline void Matrix::resize( size_t M, size_t N )
{
   // does not initialize data!

   if ( M*N != m_*n_ )
   {
      delete [] v_;
      v_ = new double[M*N];
   }
   m_ = M;
   n_ = N;
}

inline Matrix& Matrix::operator*=(const double &B)
{
    size_t mn(m_*n_);
    for (size_t i=0; i<mn; i++) v_[i] *= B;
    return *this;
}

inline Matrix& Matrix::operator+=(const Matrix &B)
{
    assert(m_==B.m_);
    assert(n_==B.n_);

    size_t mn(m_*n_);
    for (size_t i=0; i<mn; i++) v_[i] += B.v_[i];
    return *this;
}

inline Matrix& Matrix::operator-=(const Matrix &B)
{
    assert(m_==B.m_);
    assert(n_==B.n_);

    size_t mn(m_*n_);
    for (size_t i=0; i<mn; i++) v_[i] -= B.v_[i];
    return *this;
}


inline void Vector::resize( size_t M )
{
   // does not initialize data!

   if ( M != m_ )
   {
      delete [] v_;
      v_ = new double[M];
      m_ = M;
   }
}

inline Vector& Vector::operator*=(const double &B)
{
    for (size_t i=0; i<m_; i++) v_[i] *= B;
    return *this;
}

inline Vector& Vector::operator+=(const Vector &B)
{
    assert(m_==B.m_);

    for (size_t i=0; i<m_; i++) v_[i] += B.v_[i];
    return *this;
}

inline Vector& Vector::operator-=(const Vector &B)
{
    assert(m_==B.m_);

    for (size_t i=0; i<m_; i++) v_[i] -= B.v_[i];
    return *this;
}

// Matrix-Matrix and Matrix-Vector multiplications

inline Matrix mult( const Matrix &A, const Matrix &B )
{
  //  C = A * B
 
    assert(A.size2() == B.size1());
 
    size_t L = A.size1();
    size_t K = A.size2();
    size_t M = B.size2();
 
    Matrix C(L,M,0.0);
 
    for (size_t i=0; i<L; i++)
       for (size_t k=0; k<K; k++)
       {
          double *Cij( &C.v_[i*C.n_] );
          const double *Bkj( &B.v_[k*B.n_] );
          double Aik( A.v_[i*A.n_+k] );

          for (size_t j=0; j<M; j++)
             (*Cij++) += Aik * (*Bkj++);
       }

    return C;
}

inline Vector mult( const Matrix &A, const Vector &b )
{
  //  c = A * b
 
    assert(A.size2() == b.size());
 
    size_t L = A.size1();
    size_t K = b.size();
 
    Vector c(L);
 
    for (size_t i=0; i<L; i++)
    {
       const double *bk( b.v_ );
       const double *Aik( &A.v_[i*A.n_] );
       double tmp(0.0);

       for (size_t k=0; k<K; k++)
          tmp += (*Aik++) * (*bk++);

       c.v_[i] = tmp;
    }

    return c;
}

inline Matrix tmult( const Matrix &A, const Matrix &B )
{
  //       T
  //  C = A * B
 
    assert(A.size1() == B.size1());
 
    size_t L = A.size2();
    size_t K = B.size1();
    size_t M = B.size2();
 
    Matrix C(L,M,0.0);
 
    for (size_t i=0; i<L; i++)
       for (size_t k=0; k<K; k++)
       {
          double *Cij( &C.v_[i*M] );
          const double *Bkj( &B.v_[k*M] );
          double Aki( A.v_[k*L+i] );

          for (size_t j=0; j<M; j++)
             (*Cij++) += Aki * (*Bkj++);
       }

    return C;
}

inline Vector tmult( const Matrix &A, const Vector &b )
{
  //       T
  //  c = A * b
 
    assert(A.size1() == b.size());
 
    size_t L = A.size2();
    size_t K = b.size();
 
    Vector c(L,0.0);
 
    for (size_t k=0; k<K; k++)
    {
       double *ci( c.v_ );
       double *Aki( &A.v_[k*A.n_] );
       double bk( b.v_[k] );

       for (size_t i=0; i<L; i++)
          (*ci++) += (*Aki++) * bk;
//       c[i] += A[k][i] * b[k];
    }

    return c;
}

inline Matrix tmult( const Vector &a, const Matrix &B )
{
  //       T
  //  C = a * B
 
    size_t K = B.size1();
    size_t M = B.size2();
 
    assert(a.size() == K);
 
    Matrix C(1,M,0.0);
 
    for (size_t k=0; k<K; k++)
    {
       double *C0j( C.v_ );
       const double *Bkj( &B.v_[k*M] );
       double ak( a[k] );

       for (size_t j=0; j<M; j++)
          (*C0j++) += ak * (*Bkj++);
    }

    return C;
}

inline void Add_sAtBA( Matrix &C, double s, const Matrix &A, const Matrix &B )
{
  //              T
  //  C = C + s* A * B * A
 
    assert(A.size1() == B.size1());
    assert(A.size1() == B.size2());
    assert(C.size1() == A.size2());
    assert(C.size2() == A.size2());

    if ( s==0.0 ) return;
 
    size_t M = A.size1();
    size_t K = B.size1();
    size_t L = A.size2();
 
    Matrix tmp(K,L,0.0);
 
    //  tmp = B * A
    for (size_t i=0; i<K; i++)
       for (size_t k=0; k<K; k++)
       {
          double *Tij( &tmp.v_[i*L] );
          double *Akj( &A.v_[k*A.n_] );
          double Bik( B.v_[i*B.n_+k] );

          for (size_t j=0; j<L; j++)
             *Tij++ += Bik * (*Akj++);
       }

    if ( s!=1.0 )
    {
       double *T11( tmp.v_ );
       double *Tmn( &tmp.v_[K*L] );
       for (double *Tij=T11; Tij<Tmn; Tij++) *Tij *= s;
    }
 
    //  C = C + At * tmp
    for (size_t i=0; i<L; i++)
       for (size_t k=0; k<M; k++)
       {
          double *Cij( &C.v_[i*C.n_] );
          const double *Tkj( &tmp.v_[k*tmp.n_] );
          double Aki( A.v_[k*A.n_+i] );

          for (size_t j=0; j<L; j++)
             (*Cij++) += Aki * (*Tkj++);
       }
} 


//  Exercise to improve efficiency by deferred operation e.g. calculate
//  a transpose matrix multiplication without first creating a transpose
//  and perform a multiplication then. It performed however only marginally
//  better and extra parentheses were necessary in some places. Therefore
//  it was outcommented for the moment.
//
//struct Mtranspose {
//   const Matrix& m;
//
//   Mtranspose( const Matrix& mm ):m(mm){}
//   operator Matrix(){ return ~m; }
//};
//
//inline Mtranspose operator~ (const Matrix& mm )
//{
//   return Mtranspose(mm);
//}
//
//struct MtransposeMult {
//   const Matrix& mA;
//   const Matrix& mB;
//
//   MtransposeMult( const Mtranspose& A, const Matrix& B ):mA(A.m),mB(B){}
//   operator Matrix() { return tmult( mA, mB ); }
//};
//
//inline Matrix operator* (const Mtranspose& A, const Matrix& B )
//{
//   return tmult( A.m, B );
//}
//
//inline Vector operator* (const Mtranspose& A, const Vector& b )
//{
//   return tmult( A.m, b );
//}


// Definition of operators


// Matrix-Matrix and Matrix-Vector multiplication

inline Matrix operator* (const Matrix& A, const Matrix& B )
{
   return mult( A, B );
}

inline Vector operator* (const Matrix& A, const Vector& b )
{
   return mult( A, b );
}

inline Matrix operator*(const Vector &A, const Matrix &B)
{
    size_t M = A.size();
    size_t N = B.size2();

    assert(B.size1()==1);

    Matrix tmp(M,N);

    for (size_t i=0; i<M; i++)
       for (size_t j=0; j<N; j++)
          tmp[i][j] = A[i]*B[0][j];

    return tmp;
}

// transpose of Matrix and Vector

inline Matrix operator~(const Matrix &A)
{
   size_t M = A.size1();
   size_t N = A.size2();
 
   Matrix S(N,M);
   size_t i, j;
 
   for (i=0; i<M; i++)
      for (j=0; j<N; j++)
         S[j][i] = A[i][j];
 
   return S;
}

inline Matrix operator~(const Vector &A)
{
    size_t M = A.size();

    Matrix S(1,M);
    size_t i;

    for (i=0; i<M; i++)
            S[0][i] = A[i];

    return S;
}

// Scalar pre and post multiplication

inline Vector operator*(const Vector &A, const double &B)
{
    Vector tmp(A);
    tmp *= B;

    return tmp;
}

inline Vector operator*(const double &A, const Vector &B)
{
    Vector tmp(B);
    tmp *= A;

    return tmp;
}


inline Matrix operator*(const Matrix &A, const double &B)
{
    Matrix tmp(A);
    tmp *= B;

    return tmp;
}

inline Matrix operator*(const double &A, const Matrix &B)
{
    Matrix tmp(B);
    tmp *= A;

    return tmp;
}

// Matrix and Vector + and - operators

inline Vector operator+(const Vector &A, const Vector &B)
{
    assert(A.size()==B.size());

    Vector R(A);
    R += B;

    return R;
}

inline Vector operator-(const Vector &A, const Vector &B)
{
    assert(A.size()==B.size());

    Vector R(A);
    R -= B;

    return R;
}

inline Matrix operator+(const Matrix &A, const Matrix &B)
{
    assert(A.size1()==B.size1());
    assert(A.size2()==B.size2());

    Matrix R(A);
    R += B;

    return R;
}

inline Matrix operator-(const Matrix &A, const Matrix &B)
{
    assert(A.size1()==B.size1());
    assert(A.size2()==B.size2());

    Matrix R(A);
    R -= B;

    return R;
}

// specializations for mixed Matrix Vector addition and subtraction

inline Vector operator+(const Matrix &A,const Vector &B)
{
    assert(A.size1()==B.size());
    // assert(A.size2()==1); is checked in Vector constructor

    Vector tmp(A);
    tmp += B;

    return tmp;
}

inline Vector operator+(const Vector &B, const Matrix &A)
{
    assert(A.size1()==B.size());
    // assert(A.size2()==1); is checked in Vector constructor

    Vector tmp(A);
    tmp += B;

    return tmp;
}

inline Vector operator-(const Matrix &A,const Vector &B)
{
    assert(A.size1()==B.size());
    // assert(A.size2()==1); is checked in Vector constructor

    Vector tmp(A);
    tmp -= B;

    return tmp;
}

inline Vector operator-(const Vector &B, const Matrix &A)
{
    size_t M = A.size1();
    size_t N = A.size2();

    assert(M==B.size());
    assert(N==1);

    Vector tmp(B);
    size_t i;

    for (i=0; i<M; i++)
         tmp[i] -= A[i][0];

    return tmp;
}

inline double inner_product(const Vector &a, const Vector &b, double v=0)
{
   size_t N(a.size());

   assert(N==b.size());

   double *ap(a.v_);
   double *bp(b.v_);
   double *a_end( ap+N );

   for ( ap=a.v_, bp=b.v_; ap!=a_end; ap++, bp++ ) v += (*ap)*(*bp);

   return v;
}

// output stream operators

std::ostream& operator<<(std::ostream &s, const Matrix &A);
std::ostream& operator<<(std::ostream &s, const Vector &a);


} // namespace FeaTure

#endif // FTMATRIX_H
