// Copyright (c) 1999, A.H. van den Boogaard

#ifndef MATERIAL_H
#define MATERIAL_H

#include "ftmatrix.h"
#include "collect.h"

namespace FeaTure
{

enum StressType{gen,psn,pss,axi,uni};
int StressSize( StressType st );
int ExpandedStressSize( StressType st );
int get_nodal_dimension( StressType st );

class Material_Point
{
   private:
      static double time_increment;

   protected:
      Material_Point();
      static double GetTimeIncrement(){ return time_increment; }

   public:
      virtual ~Material_Point();
      virtual void finish_increment()=0;
      virtual void GetRestartData( CollectArray &data ) const=0;
      virtual void PutRestartData( const CollectArray &data )=0;

      static void SetTimeIncrement( double dtime ){ time_increment = dtime; }
};

typedef Material_Point* MatPntPtr;

class Structural_Point:public virtual Material_Point
{
   protected:
      StressType stress_type;

      Structural_Point( StressType st );
      ~Structural_Point();

      Matrix iso_lin_elas_D_mat( double Young, double Poisson, int stress_size,
                                 StressType st ) const;

      virtual Matrix get_xp_D_mat() const =0;
      virtual void ExpandEps( Vector& dstrain, Vector& ddstrain,
                              Vector& strain_bar );
      int strain_bar_size() const;
      Vector expand( Vector str ) const;
      Vector condense_vector( Vector str ) const;
      Matrix condense( Matrix str ) const;

   public:
      virtual Matrix get_D_mat() const =0;
      Vector get_D_times_iso_unit( double eps_iso=1.0 ) const;
      virtual void set_strain_increment( const Vector& strain_increment )=0;
      virtual Vector get_stress() const =0;
      virtual Vector get_xp_stress() const;
      virtual double get_eps_plast_eq( bool &has_epq ) const
              { has_epq = false; return 0; }
      void get_D_and_stress_correction( Matrix& D,
                                        Vector& stress_correction ) const;
      virtual double get_expansion_coefficient() const { return 0.0; }
};

class Thermal_Point:public virtual Material_Point
{
   protected:
      StressType dimension_type;

      Thermal_Point( StressType str );
      ~Thermal_Point();

   public:
      virtual Matrix get_conduction_matrix() const =0;
      virtual void set_flux_increment( const Vector& flux_increment )=0;
      virtual void set_potential_increment( double Tinc )=0;
      virtual double get_rho_c() const=0;
      virtual double get_accumulation() const=0;
      virtual Vector get_flux() const=0;
};

/* Material type definitions */

class Material_type
{
   private:
      double density;

   protected:
      int    nr;

   public:
      Material_type():density(0.0),nr(0){}
      Material_type( int mat_nr ):density(0.0),nr(mat_nr){}
      virtual ~Material_type(){}

      int set_density( double rho );
      double get_density() const { return density; }

      virtual Material_Point* new_Material_Point( StressType st )=0;
      virtual bool symmetric() const=0;

      int get_nr() const { return nr; }
};

class Mass_type:public Material_type
{
   private:
      double mass;

   public:
      Mass_type():Material_type(),mass(0){}
      Mass_type(int mat_nr):Material_type(mat_nr),mass(0){}
      ~Mass_type(){}

      int set_mass( double m ){ mass = m; return 0; }
      double get_mass() const { return mass; }

      Material_Point* new_Material_Point( StressType ){ return NULL; }
      bool symmetric() const { return true; }
};

class Structural_type: public virtual Material_type
{
   public:
      Structural_type();
      ~Structural_type();

      virtual bool NeedExpandStrain( StressType /*st*/ ) const { return false; }
};

class Thermal_type: public virtual Material_type
{
   public:
      Thermal_type();
      ~Thermal_type();
};

} // namespace FeaTure

#endif /* MATERIAL_H */
