#ifndef SPRSMAT_H
#define SPRSMAT_H

// Based on:
// Template Numerical Toolkit (TNT) for Linear Algebra Version 0.7
// not included in later versions)
// de-templated and further simplyfied

#include <vector>
#include <algorithm>
#include <numeric>
#include "ftmatrix.h"

namespace FeaTure
{

typedef size_t Subscript;

class ai {
  private:
    double a_;
    Subscript i_;  // i_ is only given a value at construction, so a sorted
                   // vector can not be invalidated by overwriting i_

  public:

    // constructors

    ai() : a_(0), i_(0){}
    ai(const Subscript& i, const double &a) : a_(a), i_(i){}
    ai(const ai &x) : a_(x.a_), i_(x.i_) {}
    ~ai() {}

    // read-only access 

    Subscript index() const { return i_;}
    const double & value() const { return a_;}

    // mutable access

    double & value() { return a_; }

    // comparisons
 
    bool operator==(const ai &x) const { return (a_ == x.a_ && i_ == x.i_); }
    bool operator< (const ai &x) const { return (i_ < x.i_); }
    bool operator< (const Subscript &i) const { return (i_ < i); }
};


class SparseVector
{
  /*  Allways keep the elements in the vector sorted by index */

  protected:
    typedef std::vector<ai>  Vec;

    Vec       c_;
    Subscript Dim_;

  public:

    typedef Vec::size_type              size_type;
    typedef Vec::iterator               iterator;
    typedef Vec::const_iterator         const_iterator;
    typedef Vec::reverse_iterator       reverse_iterator;
    typedef Vec::const_reverse_iterator const_reverse_iterator;
    typedef Vec::value_type             value_type;

    // generic interface to map-like structures

    iterator       begin() { return c_.begin(); }
    const_iterator begin() const { return c_.begin(); }
    iterator       end() { return c_.end(); }
    const_iterator end() const { return c_.end(); }
    reverse_iterator       rbegin() { return c_.rbegin(); }
    const_reverse_iterator rbegin() const { return c_.rbegin(); }
    reverse_iterator       rend() { return c_.rend(); }
    const_reverse_iterator rend() const { return c_.rend(); }

    // Size and shape information

    Subscript dim() const { return Dim_; }
    Subscript num_nonzeros() const {return c_.size(); }

    // Constructors

    SparseVector() : c_(), Dim_(0) {}
    SparseVector(Subscript N) : c_(), Dim_(N) {}
    SparseVector(Subscript N, Subscript nz, const double values[],
                 const Subscript indices[]): c_(), Dim_(N)
    {
        c_.reserve(nz);
        for (Subscript i=0; i<nz; i++)
            set( indices[i], values[i] );
    }

    SparseVector(const SparseVector &S) : c_(S.c_), Dim_(S.Dim_){}

    // Assignment

    SparseVector& operator= (const SparseVector& A)
    {
       c_ = A.c_;
       Dim_ = A.Dim_;
       return *this;
    }

    Subscript index(Subscript i) const { return c_[i].index(); }
    const double& value(Subscript i) const { return c_[i].value(); }
    double& value(Subscript i) { return c_[i].value(); }

    void set(Subscript i, const double& t);
    double& set(Subscript i);
    const double& get(Subscript i) const;

    const double& operator[](Subscript i) const{ return get(i); }
    double& operator[](Subscript i){ return set(i); }
};

inline Subscript vec_nonzeros( Subscript val, const SparseVector& v )
   { return val = val + v.num_nonzeros(); }


class CR_Matrix
{
    protected:
        std::vector<SparseVector> S_;

    public:

    typedef std::vector<SparseVector>::size_type              size_type;
    typedef std::vector<SparseVector>::iterator               iterator;
    typedef std::vector<SparseVector>::const_iterator         const_iterator;
    typedef std::vector<SparseVector>::reverse_iterator       reverse_iterator;
    typedef std::vector<SparseVector>::const_reverse_iterator const_reverse_iterator;
    typedef std::vector<SparseVector>::value_type             value_type;

    // generic interface to STL containers

    iterator       begin() { return S_.begin(); }
    const_iterator begin() const { return S_.begin(); }
    iterator       end() { return S_.end(); }
    const_iterator end() const { return S_.end(); }
    reverse_iterator       rbegin() { return S_.rbegin(); }
    const_reverse_iterator rbegin() const { return S_.rbegin(); }
    reverse_iterator       rend() { return S_.rend(); }
    const_reverse_iterator rend() const { return S_.rend(); }

    Subscript num_nonzeros() const
       { return std::accumulate( begin(), end(), 0, vec_nonzeros ); }

    Subscript num_rows() const { return S_.size(); }

    // Constructors

    CR_Matrix() : S_() {}//, num_cols_(0) {}  
    CR_Matrix(Subscript M, Subscript /*N*/) : S_(M) {}
    CR_Matrix( const CR_Matrix &v) : S_(v.S_) {}

    // row index is not validated in next functions (for efficiency)
    void set(Subscript i, Subscript j, const double &t) { S_[i].set(j,t); }
    double& set(Subscript i, Subscript j) { return S_[i].set(j); }

    const double& get(Subscript i, Subscript j) const { return S_[i].get(j); }
    SparseVector& operator[](Subscript i) { return S_[i]; }
    const SparseVector& operator[](Subscript i) const { return S_[i]; }
};  

// y <-- A * x 
void cr_mat_vec_mult(const CR_Matrix &A, const Vector &x, Vector *Yptr);

// Y <-- A * X 
void cr_mat_mat_mult(const CR_Matrix &A, const Matrix &x, Matrix *Yptr);

inline Vector operator*(const CR_Matrix &A, const Vector &x)
{
    Vector y;
    cr_mat_vec_mult( A, x, &y );
    return y;
}

inline Matrix operator*(const CR_Matrix &A, const Matrix &X)
{
    Matrix Y;
    cr_mat_mat_mult( A, X, &Y );
    return Y;
}

} // namespace FeaTure

#endif // SPRSMAT_H
