// Copyright (c) 1999, A.H. van den Boogaard

#include "dofs.h"

#include <iomanip>
#include <cstdlib>

using FeaTure::Direction;
using FeaTure::Dof;
using FeaTure::dof_sets;

using std::cout;
using std::ios;
using std::setw;
using std::endl;

Dof::Dof()
{
   global_dof_nr = 0;
}

void Dof::set_direction( Direction direction )
{
   dir = direction;
}

void Dof::set_dof_type( Type settype )
{
   phys_type = settype;
}

void Dof::set_dof_fixed( dof_fixed settype )
{
   fix_type = settype;
}

Direction Dof::get_direction() const
{
   return dir;
}

Dof::Type Dof::get_dof_type() const
{
   return phys_type;
}

void Dof::print() const
{
   /* Just for checking the evaluation of dofs */

   cout.setf( ios::scientific, ios::floatfield );
   cout.precision(4);

   if ( is_vector() ) cout << setw(7) << dir;
   else cout << "scalar";

   switch ( fix_type )
   {
      case fixed     : cout << " fixed,      global nr. " << setw(5)
                            << global_dof_nr << endl; break;
      case prescribed: cout << " prescribed, global nr. " << setw(5)
                            << global_dof_nr << endl; break;
      case free_dof  : cout << " free,       global nr. " << setw(5)
                            << global_dof_nr << endl;
   }
}

int Dof::get_global_dof() const
{
   return global_dof_nr;
}

void Dof::set_global_dof( int global_dof )
{
   global_dof_nr = global_dof;
}

bool Dof::is_vector() const
{
   if ( phys_type == u || phys_type == r ) return true;
   else return false;
}

bool Dof::is_scalar() const
{
   return !is_vector();
}

bool Dof::is_free() const
{
   return fix_type == free_dof;
}

bool Dof::is_fixed() const
{
   return fix_type == fixed;
}

bool Dof::is_prescribed() const
{
   return fix_type == prescribed;
}

dof_sets::dof_sets()
{
   nr_of_system_dofs = 0;
   syst_dofs = NULL;
}

/* Substitute the malloc/realloc only by a vector if you are sure that the
   vector does not reserve much too much places */

dof_sets::dof_sets( const dof_sets& in_set )
{
   nr_of_system_dofs = in_set.nr_of_system_dofs;
   if ( nr_of_system_dofs > 0 )
      syst_dofs = (dof_set*) std::malloc( nr_of_system_dofs*sizeof(dof_set) );
   else syst_dofs = NULL;

   for ( int i=0; i<nr_of_system_dofs; i++ )
      syst_dofs[i] = in_set.syst_dofs[i];
}

dof_sets::~dof_sets()
{
   if ( syst_dofs != NULL ) std::free( syst_dofs );
}

void dof_sets::add_set( FeaTure::Dof_ptr p, double f )
{
   if ( syst_dofs == NULL )
      syst_dofs = (dof_set*) std::malloc( sizeof(dof_set) );
   else
      syst_dofs = (dof_set*) std::realloc( syst_dofs, sizeof(dof_set)*(nr_of_system_dofs+1) );

   syst_dofs[nr_of_system_dofs].global_dof = p;
   syst_dofs[nr_of_system_dofs].global_contribution = f;
   nr_of_system_dofs++;
}

bool FeaTure::is_vector_dof( Dof::Type phys_type )
{
   return ( phys_type == Dof::u || phys_type == Dof::r );
}

bool FeaTure::is_scalar_dof( Dof::Type phys_type )
{
   return !is_vector_dof( phys_type );
}

void FeaTure::print_dof_type( Dof::Type type, std::ostream& out )
{
   switch ( type )
   {
      case Dof::u:  out << "u" ; break;
      case Dof::r:  out << "r" ; break;
      case Dof::t:  out << "T" ; break;
      default: out << "?" ;
   }
}

bool FeaTure::is_vector_load( Dof::Load type )
{
   return ( type == Dof::f || type == Dof::m );
}

bool FeaTure::is_scalar_load( Dof::Load type )
{
   return !is_vector_load( type );
}

void FeaTure::print_load_type( Dof::Load type, std::ostream& out )
{
   switch ( type )
   {
      case Dof::f:  out << "f" ; break;
      case Dof::m:  out << "m" ; break;
      case Dof::q:  out << "q" ; break;
      default: out << "?" ;
   }
}

Dof::Type FeaTure::conjugate_type( Dof::Load type )
{
   switch ( type )
   {
      case Dof::f:  return Dof::u;
      case Dof::m:  return Dof::r;
      case Dof::q:  return Dof::t;
      default: return Dof::u;
   }
}
