// Copyright (c) 2004, A.H. van den Boogaard

#include "eigen.h"
#include "triangular.h"

using FeaTure::Matrix;
using FeaTure::Vector;
using FeaTure::inner_product;

using std::sqrt;

namespace
{ // local namespace

void givens( double a, double b, double *c, double *s )
{
   // calculate sin(phi) and cos(phi) to get a zero matrix component
   // using Givens rotation
   // From: Matrix Computations, Golub and VanLoan, Algorithm 5.1.5

   if ( std::fabs(b)<1e-15 )
   {
      *c=1.0;
      *s=0.0;
   }
   else
   {
      if ( std::fabs(b)>std::fabs(a) )
      {
         double tau;
         tau = -a/b;
         *s = 1/sqrt(1+tau*tau);
         *c = (*s)*tau;
      }
      else
      {
         double tau;
         tau = -b/a;
         *c = 1/sqrt(1+tau*tau);
         *s = (*c)*tau;
      }
   }
}

void Givens( size_t i, size_t j, Matrix *Tp )
{
   // Perform Givens rotation in order to get T[i][j]=0

   Matrix &T(*Tp);

   size_t n( T.size1() );
   double c, s;
   givens( T[i][i], T[i][j], &c, &s );

   Matrix G(n,n,0.0);
   for ( size_t p=0; p<n; p++ ) G[p][p] = 1.0;

   G[i][i] = G[j][j] = c;
   G[i][j] = s;
   G[j][i] = -s;

   T = ~G*T*G;
}

void implicitShiftQRstep( Matrix *Tp, size_t p, size_t q, Matrix *Vp )
{
   // QR iteration with implicit shift
   // From: Matrix Computation, Golub and VanLoan, Algorithm 8.2.2

   Matrix &T(*Tp);

   size_t n( T.size1() );
   assert( T.size2()==n );
   assert( q<=n );
   assert( p<q ); // assuming (size_t)p always >= 0

   bool vectors( Vp!=0 );
   if ( vectors ) assert( Vp->size1()==n && Vp->size2()==n );

   double d=0.5*(T[q-2][q-2]-T[q-1][q-1]);
   double r, x, z;
   r= d>0 ? (d+sqrt(d*d+T[q-1][q-2]*T[q-1][q-2])):
            (d-sqrt(d*d+T[q-1][q-2]*T[q-1][q-2]));
   
   double mu=T[q-1][q-1]-T[q-1][q-2]*T[q-1][q-2]/r;
   x=T[p][p]-mu;
   z=T[p+1][p];
   size_t i;
   i = p;

   for ( size_t k=p; k<q-1; k++ )
   {
      double c, s;
      givens( x, z, &c, &s );

      // Gt*T*G ->
      double bk, zk, ap, bp, aq, bq;
      //ak = T[k][k];
      if (i>0)
         bk = T[i-1][i];
      else
         bk = 0;

      ap = T[i][i];
      if ( i<q-1)
      {
         if ( i>0 )
            zk = T[i-1][i+1];
         else
            zk = 0;

         bp = T[i][i+1];
         aq = T[i+1][i+1];

         if ( i>0 )
            T[i-1][i+1] = T[i+1][i-1] = 0;
         T[i  ][i+1] = T[i+1][i] = c*s*(ap-aq)+(c*c-s*s)*bp;
         T[i+1][i+1] = s*s*ap+c*c*aq+2*c*s*bp;

         if ( k<q-2 )
         {
            bq = T[i+1][i+2];

            T[i  ][i+2] = T[i+2][i  ] = -s*bq;
            T[i+1][i+2] = T[i+2][i+1] = c*bq;
         }
      }
      else
      {
         zk = 0;
         bp = 0;
         aq = 0;
      }

      if ( i>0 )
         T[i-1][i] = T[i][i-1] = c*bk-s*zk;
      T[i  ][i] = c*c*ap-2*c*s*bp+s*s*aq;

      if ( vectors )
      {
         // V := V*Q
         for ( size_t m=0; m<n; m++ )
         {
            Matrix &V(*Vp);
            double vi, vj;
            vi = V[m][i];
            vj = V[m][i+1];
            V[m][i]   = c*vi - s*vj;
            V[m][i+1] = s*vi + c*vj;
         }
      }

      if ( k<q-2 )
      {
         i = k+1;
         x = T[i  ][i-1];
         z = T[i+1][i-1];
      }
   }
}

void normalize( Matrix *Vp, const Matrix *Ap )
{
   // normalize each column in V such that ~Vi*A*Vi=1
   // if A==0, use ~Vi*Vi=1

   size_t n( Vp->size1() );
   size_t p( Vp->size2() );

   if ( Ap==0 )
   {
      for ( size_t i=0; i<p; i++ )
      {
         double sqlength(0);
         for ( size_t j=0; j<n; j++ )
            sqlength += (*Vp)[j][i]*(*Vp)[j][i];
   
         if (sqlength<1e-40) continue;

         for ( size_t j=0; j<n; j++ ) (*Vp)[j][i] *= 1/sqrt(sqlength);
      }
   }
   else
   {
      if ( Ap!=0 ) assert( Ap->size1()==n && Ap->size2()==n );
      Matrix U;
      U = (*Ap)*(*Vp);
      for ( size_t i=0; i<p; i++ )
      {
         double sqlength(0);
         for ( size_t j=0; j<n; j++ )
            sqlength += (*Vp)[j][i]*U[j][i];
   
         if (sqlength<1e-40) continue;

         for ( size_t j=0; j<n; j++ ) (*Vp)[j][i] *= 1/sqrt(sqlength);
      }
   }
}

} // end of local namespace

bool FeaTure::symmetric( const Matrix &A )
{
   // check wether A is symmetric;

   size_t n( A.size1() );

   if ( n!=A.size2() ) return false;

   for ( size_t i=0; i<n; i++ )
      for ( size_t j=0; j<i; j++ )
         if ( std::fabs( (A[i][j]-A[j][i])/
              (std::fabs(A[i][i])+std::fabs(A[j][j])) ) > 1e-8 ) return false;

   return true;
}

void FeaTure::householder( Matrix *Ap, Matrix *Vp )
{
   // Transform symmetric matrix A to tridiagonal form by Householder
   // transformation
   // V contains the transformation matrix
   // From: Matrix Computations, Golub and VanLoan, algorithm 8.2.1 using 5.1.1

   assert( symmetric( *Ap ) );

   Matrix &A( *Ap );

   size_t n( A.size1() );
   assert( A.size2()==n );

   bool vectors( Vp!=0 );

   if ( vectors ) assert( Vp->size1()==n && Vp->size2()==n );

   if ( n<3 ) return;
   
   for ( size_t k=0; k<n-2; k++ )
   {
      // determine Householder vector
      size_t nq( n-k-1 );
      Vector v(nq);
      for ( size_t i=0; i<nq; i++ )
         v[i] = A[k+i+1][k];
      double mu( v.norm() );
      if ( mu>1e-15 )
      {
         double beta(v[0]);
         if ( beta<0 ) mu = -mu;
         beta += mu;
         //beta += v[0]>0 ? mu : -mu;
         v *= 1/beta;
         v[0] = 1.0;
   
         // apply Householder transformation
         double sqnorm_v( inner_product(v,v) );
   
         Vector p(nq,0.0);
         for ( size_t j=0; j<nq; j++ )
         {
            double vj(v[j]);
            for ( size_t i=0; i<nq; i++ )
               p[i] += A[i+k+1][j+k+1]*vj;
         }
   
         p *= 2/sqnorm_v;
   
         Vector w(p);
         w -= v*(inner_product(p,v)/sqnorm_v);
   
         for ( size_t i=0; i<nq; i++ )
         {
            A[k+1+i][k] = A[k][k+1+i] = 0;
            for ( size_t j=0; j<nq; j++ )
               A[k+1+i][k+1+j] -= ( v[i]*w[j] + v[j]*w[i] );
         }
         A[k+1][k] = A[k][k+1] = -mu;
   
         if ( vectors )
         {
            // V := V*P
            Matrix P(n,n,0);
            for ( size_t i=0; i<n; i++ ) P[i][i] = 1.0;
   
            for ( size_t i=0; i<nq; i++ )
            {
               for ( size_t j=i; j<nq; j++ )
                  P[k+1+i][k+1+j] = P[k+1+j][k+1+i] -= 2*v[i]*v[j]/sqnorm_v;
            }
            
            (*Vp) = (*Vp)*P;
         }
      }
   }
}

void FeaTure::symmetricQR( Matrix *Ap, Matrix *Vp )
{
   // Symmetric QR algorithm
   // From: Matrix Computation, Golub and VanLoan, Algorithm 8.2.3

   double eps( 1e-15 );
   int maxiter( 10 ); // maximum iterations for 1 off-diagonal to reach eps

   size_t n( Ap->size1() );
   if ( n<2 ) return;

   if ( Vp!=0 )
   {
      Vp->resize(n,n);
      Vp->clear();
      for ( size_t i=0; i<n; i++ ) (*Vp)[i][i] = 1.0;
   }

   // tridiagonalize A by Householder transformations
   householder( Ap, Vp );
   Matrix &T(*Ap);

   size_t p, q, previous_q;

   q = n;
   previous_q = q;

   int iter(0);
   while ( q!=1 )
   //for ( int iter=0; iter<maxiter; iter++ )
   {
      for ( size_t i=0; i<q-1; i++ )
         if ( std::fabs(T[i][i+1]) <
              eps*(std::fabs(T[i][i])+std::fabs(T[i+1][i+1])) )
            T[i][i+1] = T[i+1][i] = 0.0;

      while ( q>1 && T[q-1][q-2]==0.0 ) q--;
      if ( q==1 ) break;
      if ( q==previous_q )
         iter++;
      else
         iter=0;
      if ( iter==maxiter ) break;

      p = q-2;
      while ( p>0 && T[p][p-1]!=0.0 ) p--;

      if ( q>p ) implicitShiftQRstep( &T, p, q, Vp );
   }
}

int FeaTure::SturmSequence( const Matrix &A, double shift )
{
   // Count number of sign changes in subdeterminants D(i) of a
   // sykmetric A-shift*I;
   // Golub & VanLoan, Matrix Computations, pp. 437-438

   size_t n( A.size1() );
   assert( A.size2()==n );

   if ( n==0 ) return 0;
   if ( n==1 ) return ( A[0][0]<0.0 ? 1:0 );

   Matrix B(A);

   householder(&B);

   // Now, B is a tridiagonal matrix and the sub-determinants can
   // be calculated easily.

   int a(0);

   double pr_2(1);
   double pr_1(B[0][0]-shift);

   if ( pr_1<=0 ) a++;

   for ( size_t i=1; i<n; i++ )
   {
      double pr;
      pr = (B[i][i]-shift)*pr_1 - B[i-1][i]*B[i-1][i]*pr_2;
      if ( pr==0 || pr*pr_1<0 ) a++;

      pr_2 = pr_1;
      pr_1 = pr;
   }

   return a;
}

void FeaTure::InverseIteration( const Matrix &A, Vector *xp, double *lambda )
{
   int maxiter(100);

   size_t n(A.size1());
   Vector &x( *xp );
   assert( A.size2()==n && x.size()==n );

   Matrix B(A);

   double alpha( x.norm() );
   if ( alpha>1e-15 )
      x *= 1/alpha;
   else
      x[0] = 1.0;

   double previous_mu(0);
   double delta_mu;
   delta_mu = inner_product( x, B*x );

   for ( int i=0; i<maxiter; i++ )
   {
      // perform 5 iterations without shift to approach lowest eigenvector
      if ( i>5 )
      {
         // shift for increased convergence
         for ( size_t k=0; k<n; k++ ) B[k][k] -= delta_mu;
         previous_mu += delta_mu;
      }
      x = B.inv()*x;
      alpha = x.norm();
      x *= 1/alpha;
      delta_mu = inner_product( x, B*x );
      if ( std::fabs(delta_mu) < 1e-8/(1.0+std::fabs(previous_mu)) ) break;
   }

   *lambda = delta_mu+previous_mu;
}

void FeaTure::GenEigen( Matrix *Ap, Matrix *Bp, Vector *lambda, Matrix *Vp )
{
   // determine eigenvalues and optionally eigenvectors for the
   // generalized eigenvalue problem (A-lambda*B)x = 0
   // A and B must be symmetric and B must be positive definit
   // On return A and B are destroyed
   // From: Golub and VanLoan, Matrix Computations, Algorithm 8.7.1
   
   Matrix &A( *Ap );
   Matrix &B( *Bp );

   assert( symmetric( A ) );
   assert( symmetric( B ) );
   size_t n( A.size1() );

   Matrix L( B );
   CholeskyDecomposition( &L );

   Matrix CLt(n,n);
   PostMultLowerTInv( &A, L );
   PreMultLowerInv( L, &A );

   symmetricQR( &A, Vp );
   lambda->resize(n);
   for ( size_t i=0; i<n; i++ ) (*lambda)[i] = A[i][i];

   if ( Vp!=0 ) PreMultLowerTInv( L, Vp );
}

