// Copyright (c) 1999, A.H. van den Boogaard

#include "loads.h"
#include "ftexcept.h"

#include <iomanip>

using FeaTure::Vector;
using FeaTure::SystemVector;
using FeaTure::Direction;
using FeaTure::Nodal_Loads;
using FeaTure::Nodal_Prescribed;
using FeaTure::Pressure_Loads;
using FeaTure::LoadSet;
using FeaTure::Dof;
using FeaTure::Node;

Nodal_Loads::Nodal_Loads(){}

Nodal_Loads::Nodal_Loads( int nrloads, Load_dir *lods, int nrnodes,
                          NodePtr *nods )
{
   for ( int i=0; i<nrloads; i++ ) loads.push_back(lods[i]);
   for ( int i=0; i<nrnodes; i++ ) nodes.push_back(nods[i]);
}

Nodal_Loads::~Nodal_Loads(){}

void Nodal_Loads::add_load( Dof::Load type, Direction dir, double magnitude )
{
   Load_dir new_load;
   new_load.type = type;
   new_load.dir  = dir;
   new_load.magnitude = magnitude;

   loads.push_back( new_load );
}

void Nodal_Loads::add_load( Dof::Load type, double magnitude )
{
   if ( is_vector_load( type ) )
      throw program_error( "force or moment should have a direction\n"
                           "error in Nodal_Loads::add_load( Dof::Load )" );

   Load_dir new_load;
   new_load.type = type;
   new_load.magnitude = magnitude;

   loads.push_back( new_load );
}

void Nodal_Loads::add_node( NodePtr node  )
{
   nodes.push_back( node );
}

int Nodal_Loads::get_nr_loads() const
{
   return ( loads.size() );
}

Dof::Load Nodal_Loads::get_type( int load_nr ) const
{
   return ( loads[load_nr].type );
}

Direction Nodal_Loads::get_dir(  int load_nr ) const
{
   return ( loads[load_nr].dir );
}

double Nodal_Loads::get_magnitude(  int load_nr ) const
{
   return ( loads[load_nr].magnitude );
}

int Nodal_Loads::get_nr_nodes() const
{
   return ( nodes.size() );
}

Node* Nodal_Loads::get_node( int node_nr ) const
{
   return ( nodes[node_nr] );
}

void Nodal_Loads::print( std::ostream& out ) const
{
   out.precision(4);
   out.setf( std::ios::scientific, std::ios::floatfield );

   for ( size_t i=0; i<loads.size(); i++ )
   {
      print_load_type(loads[i].type, out);
      if ( is_vector_load( loads[i].type ) )
      {
         out << "(" << std::setw(12) << loads[i].dir << ")";
      }
      out << " " << std::setw(12) << loads[i].magnitude;
   }
   out << " ";

   for ( size_t i=0; i<nodes.size(); i++ )
   {
      out << " " << std::setw(8) << nodes[i]->get_nr();
   }

   out << std::endl;
}

void Nodal_Loads::add_to_rhs( SystemVector& rhs ) const
{
   Direction dir;
   double    magnitude;
   Dof::Load type;

   for ( size_t i=0; i<loads.size(); i++ )
   {
      type = loads[i].type;
      if ( is_vector_load( type ) ) dir = loads[i].dir;
      magnitude = loads[i].magnitude;

      for ( size_t inode=0; inode<nodes.size(); inode++ )
         nodes[inode]->add_to_rhs( type, dir, magnitude, rhs );
   }
}

Nodal_Prescribed::Nodal_Prescribed(){}

Nodal_Prescribed::Nodal_Prescribed( int nrloads, Prescr_dir *lods,
                   int nrnodes, NodePtr *nods )
{
   for ( int i=0; i<nrloads; i++ ) loads.push_back(lods[i]);
   for ( int i=0; i<nrnodes; i++ ) nodes.push_back(nods[i]);
}

Nodal_Prescribed::~Nodal_Prescribed(){}

void Nodal_Prescribed::add_load( Dof::Type type, Direction dir,
                                 double magnitude )
{
   Prescr_dir new_load;
   new_load.type = type;
   new_load.dir  = dir;
   new_load.magnitude = magnitude;

   loads.push_back( new_load );
}

void Nodal_Prescribed::add_load( Dof::Type type, double magnitude )
{
   if ( is_vector_dof( type ) )
      throw program_error( "displacement or rotation should have a direction\n"
                           "error in Nodal_Prescribed::add_load( dof_type )" );

   Prescr_dir new_load;
   new_load.type = type;
   new_load.magnitude = magnitude;

   loads.push_back( new_load );
}

void Nodal_Prescribed::add_node( NodePtr node  )
{
   nodes.push_back( node );

   for ( size_t dof = 0; dof<loads.size(); dof++ )
      node->create_dof( Dof::prescribed, loads[dof].type, loads[dof].dir );
}

int Nodal_Prescribed::get_nr_loads() const
{
   return ( loads.size() );
}

Dof::Type Nodal_Prescribed::get_type( int load_nr ) const
{
   return ( loads[load_nr].type );
}

Direction Nodal_Prescribed::get_dir(  int load_nr ) const
{
   return ( loads[load_nr].dir );
}

double Nodal_Prescribed::get_magnitude(  int load_nr ) const
{
   return ( loads[load_nr].magnitude );
}

int Nodal_Prescribed::get_nr_nodes() const
{
   return ( nodes.size() );
}

Node* Nodal_Prescribed::get_node( int node_nr ) const
{
   return ( nodes[node_nr] );
}

void Nodal_Prescribed::print( std::ostream& out ) const
{
   out.precision(4);
   out.setf( std::ios::scientific, std::ios::floatfield );

   for ( size_t i=0; i<loads.size(); i++ )
   {
      print_dof_type(loads[i].type, out);
      if ( is_vector_dof( loads[i].type ) )
      {
         out << "(" << std::setw(12) << loads[i].dir << ")";
      }
      out << " " << std::setw(12) << loads[i].magnitude;
   }
   out << " ";

   for ( size_t i=0; i<nodes.size(); i++ )
   {
      out << " " << std::setw(8) << nodes[i]->get_nr();
   }

   out << std::endl;
}

void Nodal_Prescribed::add_to_prescribed( double inc_size ) const
{
   Direction dir;
   double    magnitude;
   Dof::Type type;

   for ( size_t i=0; i<loads.size(); i++ )
   {
      type = loads[i].type;
      if ( is_vector_dof( type ) ) dir = loads[i].dir;
      magnitude = loads[i].magnitude*inc_size;

      for ( size_t inode=0; inode<nodes.size(); inode++ )
         nodes[inode]->add_to_prescribed( type, dir, magnitude );
   }
}

void Pressure_Loads::add_to_vec( SystemVector& load_vec ) const
{
   for ( ConstPresIterator load = pres_list.begin();
                           load != pres_list.end(); load++ )
      (*load).elm_ptr->AddPres2Load( (*load).surface_nr,
                                     (*load).magnitude*factor, load_vec );
}

LoadSet::LoadSet():loadset_nr(0),multiplier(0.0){}

LoadSet::LoadSet( int nr ):loadset_nr(nr),multiplier(0.0){}

LoadSet::~LoadSet()
{
   size_t i;

   for ( i=0; i<nodal_loads.size(); i++ ) delete nodal_loads[i];
   for ( i=0; i<nodal_prescribed.size(); i++ ) delete nodal_prescribed[i];
   for ( i=0; i<pressure_loads.size(); i++ )   delete pressure_loads[i];
}

void LoadSet::add_load( Nodal_Loads *load )
{
   nodal_loads.push_back( load );
   return;
}

void LoadSet::add_load( Pressure_Loads *load )
{
   pressure_loads.push_back( load );
   return;
}

void LoadSet::add_load( Nodal_Prescribed *load )
{
   nodal_prescribed.push_back( load );
   return;
}

void LoadSet::print( std::ostream& out ) const
{
   size_t i;

   for ( i=0; i<nodal_loads.size(); i++ ) nodal_loads[i]->print( out );

   for ( i=0; i<nodal_prescribed.size(); i++ )
      nodal_prescribed[i]->print( out );

   return;
}

void LoadSet::add_to_prescribed( double inc_size ) const
{
   for ( size_t i=0; i<nodal_prescribed.size(); i++ )
      nodal_prescribed[i]->add_to_prescribed( inc_size );
   return;
}

void LoadSet::get_load_vector( SystemVector& load_vec ) const
{
   load_vec.clear();

   for ( size_t i=0; i<nodal_loads.size(); i++ )
      nodal_loads[i]->add_to_rhs( load_vec );               

   load_vec *= multiplier;

   for ( size_t i=0; i<pressure_loads.size(); i++ )
      pressure_loads[i]->add_to_vec( load_vec );

   return;
}

void LoadSet::set_load_size( double size )
{
   for ( size_t i=0; i<pressure_loads.size(); i++ )
         pressure_loads[i]->set_factor( size );

   add_to_prescribed( size - multiplier );
   multiplier = size;

   return;
}

void LoadSet::add_load_size( double size )
{
   for ( size_t i=0; i<pressure_loads.size(); i++ )
         pressure_loads[i]->add_factor( size );

   add_to_prescribed( size );
   multiplier += size;

   return;
}

void LoadSet::WriteRestartData( std::ostream &restart ) const
{
   restart << 'L' << get_nr() << " " << multiplier << std::endl;
}

void LoadSet::ReadRestartData( std::istream &restart )
{
   char c;
   int  n;

   restart >> c >> n;
   if ( c != 'L' || get_nr() != n ) throw file_error( "corrupt restart file" );
   restart >> multiplier;
}
