// Copyright (c) 1999, A.H. van den Boogaard

#include "material.h"
#include "ftexcept.h"

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::Material_type;
using FeaTure::Structural_type;
using FeaTure::Thermal_type;
using FeaTure::Material_Point;
using FeaTure::Structural_Point;
using FeaTure::Thermal_Point;
using FeaTure::StressType;

int Material_type::set_density( double rho )
{
   density = rho;
   if ( rho < 0.0 )
   {
      std::cerr << "Density  < 0; this is not allowed" << std::endl;
      return 1;
   }
   return 0;
}

Structural_type::Structural_type():Material_type(){}
Structural_type::~Structural_type(){}

Thermal_type::Thermal_type():Material_type(){}
Thermal_type::~Thermal_type(){}

// initialize static data
double Material_Point::time_increment = 0;

Material_Point::Material_Point(){}
Material_Point::~Material_Point(){}

Structural_Point::Structural_Point( StressType st ):stress_type(st){}
Structural_Point::~Structural_Point(){}

Vector Structural_Point::expand( Vector str ) const
{
   Vector xstr;

   switch ( stress_type )
   {
   case uni: xstr.resize(3);
             xstr(1) = str(1);
             xstr(2) = xstr(3) = 0;
             return xstr;
   case axi: return str;
   case pss:
   case psn: xstr.resize(4);
             xstr(1) = str(1);
             xstr(2) = str(2);
             xstr(3) = 0;
             xstr(4) = str(3);
             return xstr;
   case gen: return str;
   default:  throw program_error( "undefined stress state in Structural_Point::expand(.)" );
   }
}

void Structural_Point::ExpandEps( Vector& dstrain, Vector& ddstrain, Vector& strain_bar )
{
   Matrix D;
   Vector deps_xp, sigma;

   switch ( stress_type )
   {
   case uni: deps_xp = expand(dstrain);
             return;
   case pss: D = get_xp_D_mat();
             deps_xp = expand(dstrain);
             sigma = get_xp_stress();
             strain_bar(1) -= ( D(3,1)*ddstrain(1) + D(3,2)*ddstrain(2) +
                                D(3,4)*ddstrain(3) + sigma(3) ) / D(3,3);
             deps_xp(3) = strain_bar(1);
             dstrain = deps_xp;
             return;
   case psn: deps_xp = expand(dstrain);
             dstrain = deps_xp;
             return;
   default:  return;
   }
}

Vector Structural_Point::condense_vector( Vector str ) const
{
   Vector cstr;

   switch ( stress_type )
   {
   case uni: cstr.resize(1);
             cstr(1) = str(1);
             return cstr;
   case axi: return str;
   case pss:
   case psn: cstr.resize(3);
             cstr(1) = str(1);
             cstr(2) = str(2);
             cstr(3) = str(4);
             return cstr;
   case gen: return str;
   default:  throw program_error( "undefined stress state in Structural_Point::condense_vector(.)" );
   }
}

Matrix Structural_Point::condense( Matrix xp_D_mat ) const
{
   Matrix D, D12, D21;

   switch ( stress_type )
   {
   case uni: D.resize(1,1);
             D(1,1) = xp_D_mat(1,1);
             return D;
   case pss: D = xp_D_mat.without_rowcol(3);
             D12.resize(3,1);
             D21.resize(1,3);
             D12(1,1) = xp_D_mat(1,3); D21(1,1) = xp_D_mat(3,1);
             D12(2,1) = xp_D_mat(2,3); D21(1,2) = xp_D_mat(3,2);
             D12(3,1) = xp_D_mat(4,3); D21(1,3) = xp_D_mat(3,4);
             double D22inv;
             D22inv = 1/xp_D_mat(3,3);
             D -= D12*D22inv*D21;
             return D;
   case psn: D = xp_D_mat.without_rowcol(3);
             return D;
   default:  return xp_D_mat;
   }
}

Vector Structural_Point::get_D_times_iso_unit( double eps_iso ) const
{
   Vector unit;

   if ( stress_type == uni || stress_type == pss )
   {
      unit.resize( StressSize( stress_type ) );
      unit.clear();
      unit(1) = eps_iso;
      if ( stress_type == pss ) unit(2) = eps_iso;
      Matrix D = get_D_mat();
      return D*unit;
   }
   else
   {
      Matrix D = get_xp_D_mat();
      Vector unit( ExpandedStressSize( stress_type ), 0.0 );
      unit(1) = unit(2) = unit(3) = eps_iso;
      return condense_vector( D*unit );
   }
}

Vector Structural_Point::get_xp_stress() const
{
   return ( expand( get_stress() ) );
}

void Structural_Point::get_D_and_stress_correction( Matrix& D,
                                                    Vector& stress_correction ) const
{
   Matrix D_xp( get_xp_D_mat() );
   Vector sigma( get_xp_stress() );

   D = condense( D_xp );

   switch ( stress_type )
   {
   case uni: stress_correction.resize(1);
             stress_correction.clear();
             break;
   case pss: stress_correction.resize(3);
             stress_correction(1) = D_xp(1,3);
             stress_correction(2) = D_xp(2,3);
             stress_correction(3) = D_xp(4,3);
             stress_correction *= -sigma(3)/D_xp(3,3);
             break;
   default : stress_correction.resize( StressSize( stress_type ) );
             stress_correction.clear();
   }
}

int Structural_Point::strain_bar_size() const
{
   switch ( stress_type )
   {
   case uni: return 2;
   case pss: return 1;
   default : return 0;
   }
}

Matrix Structural_Point::iso_lin_elas_D_mat( double Young, double Poisson,
                                             int stress_size, StressType st ) const
{
   Matrix D(stress_size,stress_size);

   switch ( stress_size )
   {
     case 1:  D(1,1) = Young;
              break;

     case 3:  if ( st == psn )
              {
                 D(1,1) = 1-Poisson;
                 D(1,2) = Poisson;
                 D(1,3) = 0;
                 D(2,1) = Poisson;
                 D(2,2) = 1-Poisson;
                 D(2,3) = 0;
                 D(3,1) = 0;
                 D(3,2) = 0;
                 D(3,3) = (1-2*Poisson)/2;
                 D = D*(Young/(1+Poisson)/(1-2*Poisson));
              }
              else if ( st == pss )
              {
                 D(1,1) = 1;
                 D(1,2) = Poisson;
                 D(1,3) = 0;
                 D(2,1) = Poisson;
                 D(2,2) = 1;
                 D(2,3) = 0;
                 D(3,1) = 0;
                 D(3,2) = 0;
                 D(3,3) = (1-Poisson)/2;
                 D = D*(Young/(1-Poisson*Poisson));
              }
              else
                 throw program_error( "wrong stress type in iso_lin_elas_D_mat()" );

              break;

     case 4:  D(1,1) = 1-Poisson;
              D(1,2) = Poisson;
              D(1,3) = Poisson;
              D(1,4) = 0;
              D(2,1) = Poisson;
              D(2,2) = 1-Poisson;
              D(2,3) = Poisson;
              D(2,4) = 0;
              D(3,1) = Poisson;
              D(3,2) = Poisson;
              D(3,3) = 1-Poisson;
              D(3,4) = 0;
              D(4,1) = 0;
              D(4,2) = 0;
              D(4,3) = 0;
              D(4,4) = (1-2*Poisson)/2;
              D = D*(Young/(1+Poisson)/(1-2*Poisson));
              break;

     case 6:  D.clear();
              D(1,1) = 1-Poisson;
              D(1,2) = Poisson;
              D(1,3) = Poisson;

              D(2,1) = Poisson;
              D(2,2) = 1-Poisson;
              D(2,3) = Poisson;

              D(3,1) = Poisson;
              D(3,2) = Poisson;
              D(3,3) = 1-Poisson;

              D(4,4) = (1-2*Poisson)/2;
              D(5,5) = (1-2*Poisson)/2;
              D(6,6) = (1-2*Poisson)/2;

              D = D*(Young/(1+Poisson)/(1-2*Poisson));
              break;

     default:  throw program_error( "wrong number of stresses in iso_lin_elas_D_mat()" );
   }
   return D;
}

Thermal_Point::Thermal_Point( StressType str ):dimension_type( str ){}
Thermal_Point::~Thermal_Point(){}


// utilities

int FeaTure::StressSize( StressType st )
{
   switch ( st )
   {
   case uni: return 1;
   case axi: return 4;
   case pss: return 3;
   case psn: return 3;
   case gen: return 6;
   default:  throw program_error( "undefined stress state in StressSize()" );
   }
}

int FeaTure::ExpandedStressSize( StressType st )
{
   switch ( st )
   {
   case uni: return 3;
   case axi: return 4;
   case pss: return 4;
   case psn: return 4;
   case gen: return 6;
   default:  throw program_error( "undefined stress state in ExpandedStressSize()" );
   }
}

int FeaTure::get_nodal_dimension( StressType st )
{
   switch ( st )
   {
   case uni: return 1;
   case axi: return 2;
   case pss: return 2;
   case psn: return 2;
   case gen: return 3;
   default:  throw program_error( "undefined stress state in get_nodal_dimension()" );
   }
}
