// Copyright (c) 1999, 2001, 2004 A.H. van den Boogaard

#include "sysmat.h"
#include "nonsym.h"
#include "ftexcept.h"
#include "eigen.h"
#include <sstream>
#include <iomanip>
#include <cmath>

using FeaTure::SparseSysMat;
using FeaTure::SparseVector;
using FeaTure::SkySymMat;
using FeaTure::DiagonalSysMat;
using FeaTure::Matrix;
using FeaTure::Vector;

using std::ostringstream;
using std::cout;
using std::setw;
using std::endl;
using std::sqrt;

void SparseSysMat::init(size_t row, size_t col)
{
   if ( row > size || col > size || row < 1 || col < 1 )
   {
      ostringstream message;
      message << "matrix element (" << row << ',' << col
              << ") does not fit in matrix of size " << size;
      throw dimension_error( message.str() );
   }
}

void SparseSysMat::clear()
{
    size_t M = num_rows();
    size_t i;

    for (i=0; i<M; i++)  // for each row
        for ( SparseVector::iterator p=S_[i].begin(); p!=S_[i].end(); p++ )
           p->value()=0.0;

    delete Sky;
    Sky = NULL;
}

void SparseSysMat::add( const SparseSysMat &A, double s )
{
    // Add A*s to this matrix

    assert( num_rows()==A.num_rows() );
    CR_Matrix::iterator row;
    row = begin();

    // for each row
    for ( CR_Matrix::const_iterator rowA = A.begin(); rowA != A.end();
          rowA++, row++ )
        //  for each column
        for( SparseVector::const_iterator p=rowA->begin(); p!=rowA->end(); p++)
           row->set(p->index()) += s * p->value();
}

void SparseSysMat::print( char* name ) const
{
   cout << name << endl;
   cout << " i, j, value" << endl;

   size_t M = num_rows();
   size_t i;
   SparseVector::const_iterator p;

   for (i=0; i<M; i++)  // for each row
      for(p=S_[i].begin(); p!=S_[i].end(); p++)  //  for each column
         cout << i << " " << p->index() << " " << p->value() << endl;
}

void SparseSysMat::solve( const Vector& rhs, Vector& x )
{
   assert( size == num_rows() );

   size_t M = num_rows();
   size_t i;
   SparseVector::const_iterator p;

   /* If solve was called before and the matrix was not reset,
      the decomposed matrix already exsists                    */
   if ( Sky==NULL )
   {
      if ( symmetric )
         Sky = new SkySymMat( size );
      else
         Sky = new SkyNonSymMat( size );

      /* initialize Sky */
      for (i=0; i<M; i++)  // for each row
         for(p=S_[i].begin(); p!=S_[i].end(); p++)  //  for each column
            Sky->init( i+1, p->index()+1 );

      /* fill Sky */
      Sky->clear();

      for (i=0; i<M; i++)  // for each row
         for(p=S_[i].begin(); p!=S_[i].end(); p++)  //  for each column
            Sky->add( i+1, p->index()+1, p->value() );
   }

   Sky->solve( rhs, x );
}


SkySymMat::SkySymMat():SystemMat()
{
   size = 0;
   p = NULL;
   a = NULL;
   decomposed = false;
}

SkySymMat::SkySymMat( size_t n ):SystemMat()
{
   size = n;
   p = new size_t[size];
   for ( size_t i=0; i<size; i++ ) p[i] = i;
   a = NULL;
   decomposed = false;
}

SkySymMat::SkySymMat( size_t n, size_t *zeros ):SystemMat()
{
   size = n;
   p = zeros;
   create_mem();
   decomposed = false;
}

SkySymMat::SkySymMat( const SkySymMat &S ):SystemMat()
{
   assert( !(S.decomposed) );
   size = S.size;
   p = new size_t[size];
   decomposed = false;

   for ( size_t i=0; i<size; i++ ) p[i] = S.p[i];
   create_mem();

   for ( size_t i=0; i<size; i++ )
      for ( size_t j=0; j<=i-p[i]; j++ ) a[i][j] = S.a[i][j];
}

SkySymMat::~SkySymMat()
{
   if ( a != NULL )
   {
      for ( size_t i=0; i<size; i++ ) delete [] a[i];
      delete [] a;
   }
   if ( p != NULL ) delete [] p;
}

void SkySymMat::clear()
{
   if ( a==NULL ) create_mem();

   for ( size_t i=0; i<size; i++ )
      for ( size_t j=0; j<=i-p[i]; j++ ) a[i][j] = 0;

   decomposed = false;
}

void SkySymMat::init( size_t row, size_t col )
{
   if ( a!=NULL )
      throw program_error( "matrix geometry can not be updated after creation" );

   if ( row > size || col > size || row < 1 || col < 1 )
   {
      ostringstream message;
      message << "matrix element (" << row << ',' << col
              << ") does not fit in matrix of size " << size << endl;
      throw dimension_error( message.str() );
   }
   if ( col >= row ) return;

   if ( col <= p[row-1] ) p[row-1] = col-1;
   return;
}

void SkySymMat::create_mem()
{
   if ( size == 0 ) return;

   if ( a != NULL )
   {
      for ( size_t i=0; i<size; i++ ) delete [] a[i];
      delete [] a;
   }

   a = new double*[size];
   for ( size_t i=0; i<size; i++ ) a[i] = new double[i-p[i]+1];
}

void SkySymMat::add( size_t row, size_t col, double value )
{
   if ( a==NULL ) create_mem();

   if ( decomposed )
      throw program_error( "can not add matrix component to a decomposed matrix" );

   if ( row > size || col > size || row < 1 || col < 1 )
   {
      ostringstream message;
      message << "matrix element (" << row << ',' << col
              << ") does not fit in matrix of size " << size;
      throw dimension_error( message.str() );
   }
   if ( col > row ) return;

   a[row-1][col-1-p[row-1]] += value;
}

void SkySymMat::print( char *name ) const
{
   double value;

   if ( decomposed ) cout << "Decomposed ";
   cout << name << endl << endl;

   cout.precision(3);
   cout.setf( std::ios::scientific, std::ios::floatfield );

   for ( size_t row = 0; row < size && row < 10 ; row++ )
   {
      for ( size_t col = 0; col <= row; col++ )
      {
         if ( col < p[row] ) value = 0;
         else value = a[row][col-p[row]];
         cout << setw(10) << value;
      }
      cout << endl;
   }
}

void SkySymMat::solve( const Vector& rhs, Vector& x )
{
   size_t nr_zeros;

   if ( size == 0 ) return;

   assert( a != NULL );

   if ( !decomposed )
   {
      nr_zeros = factor();

      if ( nr_zeros > 0 )
         cout << "there are " << nr_zeros << " (almost) zero pivots found!"
              << endl;
      decomposed = true;
   }

   x.resize( size );

   substitute( x, rhs );

   return;
}

/*
   Factoring and substituting routines for symmetric skyline matrix,
   based on Computers and Structures Vol. 52, No. 4, pp. 743-748, 1994
   O. Hededal and S. Krenk, A Profile Solver in C for Finite Element
   Equations.

   Adapted to catch large ratio of pivots.
   Adapted by using pointer arithmetic, speeded up by a factor of 3.2 !!!
   (using pgcc-2.91.66 with O2 flag)
*/

int SkySymMat::factor()
/* factorize matrix

   Factors profile matrix [A] defined in the lower triangular
   part into off-diagonal factors [L] and the diagonal [D]:
                [A] = [L][D][L]T
   Rows and columns identified by f=1 are left untouched.

   members:
      size      = Dimension of the matrix a[][].
      a[i][j]   = Lower profile matrix to be factored.
                  i=0 to n-1, j=p[i] to i+1.
      p[i]      = Number of leading zeroes in row i.

   output:
      a[i][j]   = i>j: Profile factor [L] below diagonal.
                = i=j: Diagonal factor [D] on diagonal.
      status    = No. of tiny diagonal elements
*/
{
   const double tiny( 1e-20 ); /* to avoid numerical instability */

   size_t i,j,k;
   int status=0;
   double u, maxpiv(0), minpiv(0), abspiv;

   for ( i=0; i<size; i++ )
   {
      double *A_i( a[i] );

      for ( j=p[i]; j<i; j++ )
      {
         double A_ij( A_i[j-p[i]] );

         size_t start( std::max(p[i],p[j]) );

         double *A_ik( &A_i[start-p[i]] );
         double *A_jk( &a[j][start-p[j]] );

         for ( k=start; k<j; k++ )
            A_ij -= (*A_jk++)*(*A_ik++);

         A_i[j-p[i]] = A_ij;
      }

      double *A_ij = A_i;
      double A_ii = A_i[i-p[i]];

      for ( j=p[i]; j<i; j++ )
      {
         u = *A_ij;
         *A_ij /= a[j][j-p[j]];
         A_ii -= (*A_ij++)*u;
      }

      abspiv = std::fabs(A_ii);
      if ( abspiv < tiny )
      {
         A_ii = tiny;
         status++;
      }

      A_i[i-p[i]] = A_ii;

      if ( i==0 ) maxpiv = minpiv = abspiv;
      else if ( abspiv > maxpiv ) maxpiv = abspiv;
      else if ( abspiv < minpiv ) minpiv = abspiv;
   }

   if ( minpiv / maxpiv < 1.e-10 ) status = std::max( status, 1 );

   return status;
}

void SkySymMat::substitute( Vector &x, const Vector &y ) const
/*  solve:
     Solves a symmetric system of equations, when the profile
     matrix [A] has been factored in the form
               [L][D][L]T {x} = {y}
     [L] is a lower profile matrix and [D] is a diagonal matrix.

     members
       size     = Dimension of equation system.
       a[i][j]  = i>j: Off-diagonal factor [L].
                = i=j: Diagonal factor [D].
                  i=0 to n-1, j=p[i]+1 to i.
       p[i]     = Number of leading zeroes in a row.

     input:
       x[i]     = System degrees of freedom.
       y[i]     = Load vector.

     output:
       x[i]     = Degrees of freedom.
       y[i]     = Loads.
*/
{
   size_t i,j;

   for ( i=0; i<size; i++ )
      x[i] = y[i];

   for ( i=1; i<size; i++ )
   {
      double *A_ij( a[i] );

      for ( j=p[i]; j<i; j++ )
         x[i] -= (*A_ij++)*x[j];
   }

   for ( i=0; i<size; i++ )
      x[i] /= a[i][i-p[i]];

   for ( i=size-1; i>0; i-- )
   {
      double *A_ij( a[i] );

      for ( j=p[i]; j<i; j++ )
         x[j] -= (*A_ij++)*x[i];
   }
}

DiagonalSysMat::DiagonalSysMat():SystemMat()
{
   size = 0;
   a = NULL;
   inverted = false;
}

DiagonalSysMat::DiagonalSysMat( size_t n ):SystemMat()
{
   size = n;
   a = new double[size];
   inverted = false;
}

DiagonalSysMat::DiagonalSysMat( const DiagonalSysMat &D ):SystemMat()
{
   assert( !(D.inverted) );
   size = D.size;
   a = new double[size];
   inverted = false;
   for ( size_t i=0; i<size; i++ ) a[i] = D.a[i];
}

DiagonalSysMat::~DiagonalSysMat()
{
   delete [] a;
}

void DiagonalSysMat::clear()
{
   for ( size_t i=0; i<size; i++ ) a[i] = 0;

   inverted = false;
}

void DiagonalSysMat::init( size_t row, size_t col )
{
   if ( row != col || row > size || row < 1 )
   {
      ostringstream message;
      message << "matrix element (" << row << ',' << col
              << ") does not fit in diagonal matrix of size " << size;
      throw dimension_error( message.str() );
   }

   return;
}

void DiagonalSysMat::add( size_t row, size_t col, double value )
{
   if ( inverted )
      throw program_error( "can not add matrix component to an inverted matrix");

   if ( row != col || row > size || row < 1 )
   {
      ostringstream message;
      message << "matrix element (" << row << ',' << col
              << ") does not fit in diagonal matrix of size " << size;
      throw dimension_error( message.str() );
   }

   a[row-1] += value;
}

void DiagonalSysMat::print( char *name ) const
{
   double value;

   if ( inverted ) cout << "Inverted ";
   cout << name << endl << endl;

   cout.precision(3);
   cout.setf( std::ios::scientific, std::ios::floatfield );

   for ( size_t row = 0; row < size && row < 10 ; row++ )
   {
      for ( size_t col = 0; col <= row; col++ )
      {
         if ( col < row ) value = 0;
         else value = a[row];
         cout << setw(10) << value;
      }
      cout << endl;
   }
}

void DiagonalSysMat::solve( const Vector& rhs, Vector& x )
{
   size_t nr_zeros;

   if ( !inverted )
   {
      nr_zeros = 0;
      for ( size_t i=0; i<size; i++ )
      {
         if ( std::fabs(a[i]) < 1.e-12 )
            nr_zeros++;
         else
            a[i] = 1.0/a[i];
      }

      if ( nr_zeros > 0 )
      {
         ostringstream message;
         message << "there are " << nr_zeros << " (almost) zero pivots found!";
         throw matrix_error( message.str() );
      }

      inverted = true;
   }

   x.resize( size );

   for ( size_t i=0; i<size; i++ ) x[i] = rhs[i]*a[i];

   return;
}

void DiagonalSysMat::mult( const Vector& x, Vector* bp ) const
{
   size_t n( x.size() );
   assert( n==size );

   Vector &b( *bp );

   b.resize(n);

   for ( size_t i=0; i<size; i++ ) b[i] = a[i]*x[i];
}

void DiagonalSysMat::selfmult( double s )
{
   for ( size_t i=0; i<size; i++ ) a[i] *= s;
}

class SizeDoublePair : public std::pair<size_t,double>
{
   public:
      bool operator< (const SizeDoublePair& c) const { return second<c.second; }
};

void FeaTure::subspace_iteration( SparseSysMat &K, const SparseSysMat &M,
                                  Vector *lambda, Matrix *Xp, bool init )
{
   // perform subspace iteration for the generalized eigenvalue problem
   // Kx=lambda*Mx
   // X should be initialized to p trial vectors that will iterate towards the
   // eigenvectors, or if init is true initialization is performed here
   // From: Introduction to finite element vibration analysis, M. Petyt, p. 378

   int maxiter(100);
   double eps( 1e-10 );

   Matrix &X( *Xp );

   size_t n( X.size1() );
   size_t p( X.size2() );

   assert( p<=n );
   assert( n==K.num_rows() );

   lambda->resize(p);

   if ( p==0 ) return;

   Vector x(n);
   Vector rhs;

   Matrix Kr, Mr;
   Matrix Xn(n,p);

   if ( init )
   {
      X.clear();

      // first vector is composed of diagonal elements of M
      for ( size_t i=0; i<n; i++ ) X[i][0] = M[i][i];

      // rest of vectors are unit vectors for lowest Kii/Mii values
      std::vector<SizeDoublePair>  k_over_m(n);
      for ( size_t i=0; i<n; i++ )
      {
         k_over_m[i].first = i;
         k_over_m[i].second = K[i][i]/M[i][i];
      }
      std::vector<SizeDoublePair> lowest(p);
      std::partial_sort_copy( k_over_m.begin(), k_over_m.end(),
                              lowest.begin(), lowest.end() );

      for ( size_t j=1; j<p; j++ )
      {
         size_t i( lowest[j-1].first );
         X[i][j] = 1.0;
      }
   }

   int iter;
   for ( iter=0; iter<maxiter; iter++ )
   {
      for ( size_t k=0; k<p; k++ )
      {
         if ( iter==0 && init )
         {
            rhs.resize(n);
            for ( size_t i=0; i<n; i++ ) rhs[i] = X[i][k];
         }
         else
         {
            for ( size_t i=0; i<n; i++ ) x[i] = X[i][k];
            rhs = M*x;
         }
         K.solve(rhs,x);
         double alpha( sqrt(inner_product( x, M*x )) );
         x *= 1.0/alpha;
         for ( size_t i=0; i<n; i++ ) Xn[i][k] = x[i];
      }

      Kr = ~Xn*(K*Xn);
      Mr = ~Xn*(M*Xn);

      // check orthogonality

      double diagK(0), diagM(0), offdiagK(0), offdiagM(0);

      for ( size_t i=0; i<p; i++ )
      {
         diagK += std::fabs(Kr[i][i]);
         diagM += std::fabs(Mr[i][i]);
         for ( size_t j=0; j<i; j++ )
         {
            offdiagK += std::fabs(Kr[i][j]);
            offdiagM += std::fabs(Mr[i][j]);
         }
      }

      if ( offdiagK/p < eps*diagK && offdiagM/p < eps*diagM ) break;

      Matrix Q;
      GenEigen( &Kr, &Mr, lambda, &Q );

      X = Xn*Q;
   }

   cout << "subspace iteration converged in " << iter << " iterations" << endl;
}
